/* packet-rtpmidi.c
 *
 * Routines for RFC 4695 RTP-MIDI dissection
 * Copyright 2006-2010, Tobias Erichsen <t.erichsen@gmx.de>
 *
 * $Id$
 *
 * Wireshark - Network traffic analyzer
 * By Gerald Combs <gerald@wireshark.org>'
 * Copyright 1998 Gerald Combs
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * RTP-MIDI ( RFC-4695) is a payload type for transmitting MIDI-data over RTP-
 * packets.  This payload type can be used on an RTP-conversation that has been
 * established either by means of a session establishment-protocol like SIP or
 * something more lightweight like the Apple network-midi session establishment.
 * Using RTP-Midi payload is also possible without session-establishment.
 *
 * Here are some links:
 *
 * http://www.cs.berkeley.edu/~lazzaro/rtpmidi/
 * http://www.faqs.org/rfcs/rfc4695.html
 * http://www.midi.org/
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <glib.h>
#include <epan/packet.h>
#include <epan/prefs.h>

#include <stdio.h>
#include <string.h>
#include "packet-rtp-midi.h"
#include <epan/conversation.h>
#include <epan/tap.h>

/* rtp_midi_payload_type_value is the value used globally
 * to set the appropriate payload type
 * saved_pt_value is a temporary place to save the value
 * so we can properly reinitialize when the settings
 * get changed
 *
 * Since the AppleMIDI session control protocol is the
 * only currently known protocol that is being in use to
 * set up a RTP-MIDI session and AppleMIDI always uses
 * a payload-type of 97, we default to this value here.
 * This - and the whole mechanism could be changed later
 * if an implementation is encountered that is using SIP/SDP
 * for example to setup an RTP-MIDI-session
 */
static guint rtp_midi_payload_type_value	= 97;
static guint saved_payload_type_value;


/* RTP Event Fields */

static int proto_rtp_midi			= -1;




void proto_reg_handoff_rtp_midi( void );


/*
 * This decodes the delta-time before a MIDI-command
 */
static int
decodetime(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned int offset, unsigned int cmd_len)
{
	guint8		octet;
	unsigned int	consumed;
	unsigned int    i;
	unsigned long   deltatime;

	consumed  = 0;
	deltatime = 0;


	/* RTP-MIDI deltatime is "compressed" using only the necessary amount of octets */
	for ( i=0; i < 4; i++ ) {

		if ( !cmd_len ) {
			return -1;
		}

		octet = tvb_get_guint8( tvb, offset + consumed );
		deltatime = ( deltatime << 7 ) | ( octet & RTP_MIDI_DELTA_TIME_OCTET_MASK );
		consumed++;

		if ( ( octet & RTP_MIDI_DELTA_TIME_EXTENSION ) == 0 ) {
			break;
		}
	}

	switch (consumed) {
		case 1:
		    proto_tree_add_uint ( tree, hf_rtp_midi_deltatime1, tvb, offset, consumed, deltatime );
			break;
		case 2:
		    proto_tree_add_uint ( tree, hf_rtp_midi_deltatime2, tvb, offset, consumed, deltatime );
			break;
		case 3:
		    proto_tree_add_uint ( tree, hf_rtp_midi_deltatime3, tvb, offset, consumed, deltatime );
			break;
		case 4:
		    proto_tree_add_uint ( tree, hf_rtp_midi_deltatime4, tvb, offset, consumed, deltatime );
			break;

	}

	return consumed;
}



/*
 * Here a Note-Off command is decoded.
 */
static int
decode_note_off(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned int cmd_count, unsigned int offset, unsigned int cmd_len, guint8 status, unsigned int rsoffset, gboolean using_rs ) {

	guint8		 note		= 0;
	guint8		 velocity	= 0;
	const gchar	*status_str	= NULL;
	const gchar	*note_str	= NULL;
	proto_item	*command_item	= NULL;
	proto_tree	*command_tree	= NULL;

	status_str = val_to_str( status >> 4, rtp_midi_channel_status, "" );

	/* broken: we have no further data */
	if ( !cmd_len ) {
		/* this case should never happen */
		if ( using_rs ) {
			command_item = proto_tree_add_text( tree, tvb, offset, 0, "TRUNCATED: %s (c=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1 );
			command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, rsoffset, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, rsoffset, 1, FALSE );

		} else {
			command_item = proto_tree_add_text( tree, tvb, offset - 1, 1, "TRUNCATED: %s (c=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1);
			command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, offset - 1, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, offset - 1, 1, FALSE );
			}
		return -1;
	}

	note = tvb_get_guint8( tvb, offset );

	/* seems to be an aborted MIDI-command */
	if ( note & RTP_MIDI_COMMAND_STATUS_FLAG ) {
		/* this case should never happen */
		if ( using_rs ) {
			command_item = proto_tree_add_text( tree, tvb, offset, 0, "ABORTED: %s (c=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1 );
			command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, rsoffset, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, rsoffset, 1, FALSE );

		} else {
			command_item = proto_tree_add_text( tree, tvb, offset - 1, 1, "ABORTED: %s (c=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1 );
			command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, offset - 1, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, offset - 1, 1, FALSE );
			}
		return -1;
	}

	note_str = val_to_str( note, rtp_midi_note_values, "" );

	/* broken: we have only one further octet */
	if ( cmd_len < 2 ) {
		if ( using_rs ) {
			command_item = proto_tree_add_text( tree, tvb, offset, 1, "TRUNCATED: %s (c=%d, n=%s)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1, note_str );
			command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, rsoffset, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, rsoffset, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_note, tvb, offset, 1, FALSE );

		} else {
			command_item = proto_tree_add_text( tree, tvb, offset - 1, 2, "TRUNCATED: %s (c=%d, n=%s)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1, note_str );
			command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, offset - 1, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, offset - 1, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_note, tvb, offset, 1, FALSE );
			}
		return -1;
	}

	velocity = tvb_get_guint8( tvb, offset + 1 );

	/* seems to be an aborted MIDI-command */
	if ( velocity  & RTP_MIDI_COMMAND_STATUS_FLAG ) {
		if ( using_rs ) {
			command_item = proto_tree_add_text( tree, tvb, offset, 1, "ABORTED: %s (c=%d, n=%s)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1, note_str );
			command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, rsoffset, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, rsoffset, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_note, tvb, offset, 1, FALSE );

		} else {
			command_item = proto_tree_add_text( tree, tvb, offset - 1, 2, "ABORTED: %s (c=%d, n=%s)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1, note_str );
			command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, offset - 1, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, offset - 1, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_note, tvb, offset, 1, FALSE );
			}
		return -1;
	}


	if ( using_rs ) {
		command_item = proto_tree_add_text( tree, tvb, offset, 2, "%s (c=%d, n=%s, v=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1, note_str, velocity );
		command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
		proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, rsoffset, 1, FALSE );
		proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, rsoffset, 1, FALSE );
		proto_tree_add_item ( command_tree, hf_rtp_midi_note, tvb, offset, 1, FALSE );
		proto_tree_add_item ( command_tree, hf_rtp_midi_velocity, tvb, offset + 1, 1, FALSE );
	} else {
		command_item = proto_tree_add_text(tree, tvb, offset - 1, 3, "%s (c=%d, n=%s, v=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1, note_str, velocity );
		command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
		proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, offset - 1, 1, FALSE );
		proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, offset - 1, 1, FALSE );
		proto_tree_add_item ( command_tree, hf_rtp_midi_note, tvb, offset, 1, FALSE );
		proto_tree_add_item ( command_tree, hf_rtp_midi_velocity, tvb, offset + 1, 1, FALSE );
	}

	if ( cmd_count ) {
		col_append_fstr(pinfo->cinfo, COL_INFO,", %s (c=%d, n=%s, v=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1, note_str, velocity );
	} else {
		col_append_fstr(pinfo->cinfo, COL_INFO, "%s (c=%d, n=%s, v=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1, note_str, velocity );
	}
	return 2;
}


/*
 * Here a Note-On command is decoded.
 */
static int
decode_note_on(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned int cmd_count, unsigned int offset, unsigned int cmd_len, guint8 status, unsigned int rsoffset, gboolean using_rs ) {

	guint8		 note		= 0;
	guint8		 velocity	= 0;
	const gchar	*status_str	= NULL;
	const gchar	*note_str	= NULL;
	proto_item	*command_item	= NULL;
	proto_tree	*command_tree	= NULL;

	status_str = val_to_str( status >> 4, rtp_midi_channel_status, "" );

	/* broken: we have no further data */
	if ( !cmd_len ) {
		/*  this should never happen! */
		if ( using_rs ) {
			command_item = proto_tree_add_text( tree, tvb, offset, 0, "TRUNCATED: %s (c=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1 );
			command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, rsoffset, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, rsoffset, 1, FALSE );

		} else {
			command_item = proto_tree_add_text( tree, tvb, offset - 1, 1, "TRUNCATED: %s (c=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1);
			command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, offset - 1, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, offset - 1, 1, FALSE );
			}
		return -1;
	}

	note = tvb_get_guint8( tvb, offset );

	/* seems to be an aborted MIDI-command */
	if ( note & RTP_MIDI_COMMAND_STATUS_FLAG ) {
		/* this should never happethis should never happen */
		if ( using_rs ) {
			command_item = proto_tree_add_text( tree, tvb, offset, 0, "ABORTED: %s (c=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1 );
			command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, rsoffset, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, rsoffset, 1, FALSE );

		} else {
			command_item = proto_tree_add_text( tree, tvb, offset - 1, 1, "ABORTED: %s (c=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1 );
			command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, offset - 1, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, offset - 1, 1, FALSE );
			}
		return -1;
	}

	note_str = val_to_str( note, rtp_midi_note_values, "" );

	/* broken: we have only one further octet */
	if ( cmd_len < 2 ) {
		if ( using_rs ) {
			command_item = proto_tree_add_text( tree, tvb, offset, 1, "TRUNCATED: %s (c=%d, n=%s)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1, note_str );
			command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, rsoffset, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, rsoffset, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_note, tvb, offset, 1, FALSE );

		} else {
			command_item = proto_tree_add_text( tree, tvb, offset - 1, 2, "TRUNCATED: %s (c=%d, n=%s)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1, note_str );
			command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, offset - 1, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, offset - 1, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_note, tvb, offset, 1, FALSE );
			}
		return -1;
	}

	velocity = tvb_get_guint8( tvb, offset + 1 );

	/* seems to be an aborted MIDI-command */
	if ( velocity  & RTP_MIDI_COMMAND_STATUS_FLAG ) {
		if ( using_rs ) {
			command_item = proto_tree_add_text( tree, tvb, offset, 1, "ABORTED: %s (c=%d, n=%s)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1, note_str );
			command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, rsoffset, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, rsoffset, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_note, tvb, offset, 1, FALSE );

		} else {
			command_item = proto_tree_add_text( tree, tvb, offset - 1, 2, "ABORTED: %s (c=%d, n=%s)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1, note_str );
			command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, offset - 1, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, offset - 1, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_note, tvb, offset, 1, FALSE );
			}
		return -1;
	}

	/* special case velocity=0 for Note-On means Note-Off (to preserve running-status!) */
	if ( velocity == 0 ) {
		status_str = rtp_midi_channel_status_special_off;
	}

	if ( using_rs ) {
		command_item = proto_tree_add_text( tree, tvb, offset, 2, "%s (c=%d, n=%s, v=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1, note_str, velocity );
		command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
		proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, rsoffset, 1, FALSE );
		proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, rsoffset, 1, FALSE );
		proto_tree_add_item ( command_tree, hf_rtp_midi_note, tvb, offset, 1, FALSE );
		proto_tree_add_item ( command_tree, hf_rtp_midi_velocity, tvb, offset + 1, 1, FALSE );
	} else {
		command_item = proto_tree_add_text(tree, tvb, offset - 1, 3, "%s (c=%d, n=%s, v=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1, note_str, velocity );
		command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
		proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, offset - 1, 1, FALSE );
		proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, offset - 1, 1, FALSE );
		proto_tree_add_item ( command_tree, hf_rtp_midi_note, tvb, offset, 1, FALSE );
		proto_tree_add_item ( command_tree, hf_rtp_midi_velocity, tvb, offset + 1, 1, 0);
	}

	if ( cmd_count ) {
		col_append_fstr(pinfo->cinfo, COL_INFO,", %s (c=%d, n=%s, v=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1, note_str, velocity );
	} else {
		col_append_fstr(pinfo->cinfo, COL_INFO, "%s (c=%d, n=%s, v=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1, note_str, velocity );
	}
	return 2;
}




/*
 * Here polyphonic aftertouch is decoded.
 */
static int
decode_poly_pressure(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned int cmd_count, unsigned int offset, unsigned int cmd_len, guint8 status, unsigned int rsoffset, gboolean using_rs ) {

	guint8		 note		= 0;
	guint8		 pressure	= 0;
	const gchar	*status_str	= NULL;
	const gchar	*note_str	= NULL;
	proto_item	*command_item	= NULL;
	proto_tree	*command_tree	= NULL;

	status_str = val_to_str( status >> 4, rtp_midi_channel_status, "" );

	/* broken: we have no further data */
	if ( !cmd_len ) {
		/* this should never happen */
		if ( using_rs ) {
			command_item = proto_tree_add_text( tree, tvb, offset, 0, "TRUNCATED: %s (c=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1 );
			command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, rsoffset, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, rsoffset, 1, FALSE );

		} else {
			command_item = proto_tree_add_text( tree, tvb, offset - 1, 1, "TRUNCATED: %s (c=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1);
			command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, offset - 1, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, offset - 1, 1, FALSE );
			}
		return -1;
	}

	note = tvb_get_guint8( tvb, offset );

	/* seems to be an aborted MIDI-command */
	if ( note & RTP_MIDI_COMMAND_STATUS_FLAG ) {
		/* this should never happen */
		if ( using_rs ) {
			command_item = proto_tree_add_text( tree, tvb, offset, 0, "ABORTED: %s (c=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1 );
			command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, rsoffset, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, rsoffset, 1, FALSE );

		} else {
			command_item = proto_tree_add_text( tree, tvb, offset - 1, 1, "ABORTED: %s (c=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1 );
			command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, offset - 1, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, offset - 1, 1, FALSE );
			}
		return -1;
	}

	note_str = val_to_str( note, rtp_midi_note_values, "" );

	/* broken: we have only one further octet */
	if ( cmd_len < 2 ) {
		if ( using_rs ) {
			command_item = proto_tree_add_text( tree, tvb, offset, 1, "TRUNCATED: %s (c=%d, n=%s)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1, note_str );
			command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, rsoffset, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, rsoffset, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_note, tvb, offset, 1, FALSE );

		} else {
			command_item = proto_tree_add_text( tree, tvb, offset - 1, 2, "TRUNCATED: %s (c=%d, n=%s)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1, note_str );
			command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, offset - 1, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, offset - 1, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_note, tvb, offset, 1, FALSE );
			}
		return -1;
	}

	pressure = tvb_get_guint8( tvb, offset + 1 );

	/* seems to be an aborted MIDI-command */
	if ( pressure  & RTP_MIDI_COMMAND_STATUS_FLAG ) {
		if ( using_rs ) {
			command_item = proto_tree_add_text( tree, tvb, offset, 1, "ABORTED: %s (c=%d, n=%s)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1, note_str );
			command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, rsoffset, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, rsoffset, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_note, tvb, offset, 1, FALSE );

		} else {
			command_item = proto_tree_add_text( tree, tvb, offset - 1, 2, "ABORTED: %s (c=%d, n=%s)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1, note_str );
			command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, offset - 1, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, offset - 1, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_note, tvb, offset, 1, FALSE );
			}
		return -1;
	}


	if ( using_rs ) {
		command_item = proto_tree_add_text( tree, tvb, offset, 2, "%s (c=%d, n=%s, p=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1, note_str, pressure );
		command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
		proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, rsoffset, 1, FALSE );
		proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, rsoffset, 1, FALSE );
		proto_tree_add_item ( command_tree, hf_rtp_midi_note, tvb, offset, 1, FALSE );
		proto_tree_add_item ( command_tree, hf_rtp_midi_pressure, tvb, offset + 1, 1, FALSE );
	} else {
		command_item = proto_tree_add_text(tree, tvb, offset - 1, 3, "%s (c=%d, n=%s, p=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1, note_str, pressure );
		command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
		proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, offset - 1, 1, FALSE );
		proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, offset - 1, 1, FALSE );
		proto_tree_add_item ( command_tree, hf_rtp_midi_note, tvb, offset, 1, FALSE );
		proto_tree_add_item ( command_tree, hf_rtp_midi_pressure, tvb, offset + 1, 1, 0);
	}

	if ( cmd_count ) {
		col_append_fstr(pinfo->cinfo, COL_INFO,", %s (c=%d, n=%s, p=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1, note_str, pressure );
	} else {
		col_append_fstr(pinfo->cinfo, COL_INFO, "%s (c=%d, n=%s, p=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1, note_str, pressure );
	}
	return 2;
}


/*
 * Here channel aftertouch is decoded.
 */
static int
decode_channel_pressure(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned int cmd_count, unsigned int offset, unsigned int cmd_len, guint8 status, unsigned int rsoffset, gboolean using_rs ) {

	guint8		 pressure	= 0;
	const gchar	*status_str	= NULL;
	proto_item	*command_item	= NULL;
	proto_tree	*command_tree	= NULL;

	status_str = val_to_str( status >> 4, rtp_midi_channel_status, "" );

	/* broken: we have no further data */
	if ( !cmd_len ) {
		/* this should never happen */
		if ( using_rs ) {
			command_item = proto_tree_add_text( tree, tvb, offset, 0, "TRUNCATED: %s (c=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1 );
			command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, rsoffset, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, rsoffset, 1, FALSE );

		} else {
			command_item = proto_tree_add_text( tree, tvb, offset - 1, 1, "TRUNCATED: %s (c=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1);
			command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, offset - 1, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, offset - 1, 1, FALSE );
			}
		return -1;
	}

	pressure = tvb_get_guint8( tvb, offset );

	/* seems to be an aborted MIDI-command */
	if ( pressure  & RTP_MIDI_COMMAND_STATUS_FLAG ) {
		if ( using_rs ) {
			/* this should never happen */
			command_item = proto_tree_add_text( tree, tvb, offset, 0, "ABORTED: %s (c=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1 );
			command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, rsoffset, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, rsoffset, 1, FALSE );

		} else {
			command_item = proto_tree_add_text( tree, tvb, offset - 1, 1, "ABORTED: %s (c=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1);
			command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, offset - 1, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, offset - 1, 1, FALSE );
			}
		return -1;
	}


	if ( using_rs ) {
		command_item = proto_tree_add_text( tree, tvb, offset, 1, "%s (c=%d, p=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1, pressure );
		command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
		proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, rsoffset, 1, FALSE );
		proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, rsoffset, 1, FALSE );
		proto_tree_add_item ( command_tree, hf_rtp_midi_channel_pressure, tvb, offset, 1, FALSE );
	} else {
		command_item = proto_tree_add_text(tree, tvb, offset - 1, 2, "%s (c=%d, p=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1, pressure );
		command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
		proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, offset - 1, 1, FALSE );
		proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, offset - 1, 1, FALSE );
		proto_tree_add_item ( command_tree, hf_rtp_midi_channel_pressure, tvb, offset, 1, 0);
	}

	if ( cmd_count ) {
		col_append_fstr(pinfo->cinfo, COL_INFO,", %s (c=%d, p=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1, pressure );
	} else {
		col_append_fstr(pinfo->cinfo, COL_INFO, "%s (c=%d, p=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1, pressure );
	}
	return 1;
}



/*
 * Here pitch-bend is decoded.
 */
static int
decode_pitch_bend_change(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned int cmd_count, unsigned int offset, unsigned int cmd_len, guint8 status, unsigned int rsoffset, gboolean using_rs ) {

	guint8		 octet1		= 0;
	guint8		 octet2		= 0;
	guint8		 pitch		= 0;
	const gchar	*status_str	= NULL;
	const gchar	*note_str	= NULL;
	proto_item	*command_item   = NULL;
	proto_tree	*command_tree   = NULL;

	status_str = val_to_str( status >> 4, rtp_midi_channel_status, "" );

	/* broken: we have no further data */
	if ( !cmd_len ) {
		/* this should never happen */
		if ( using_rs ) {
			command_item = proto_tree_add_text( tree, tvb, offset, 0, "TRUNCATED: %s (c=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1 );
			command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, rsoffset, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, rsoffset, 1, FALSE );

		} else {
			command_item = proto_tree_add_text( tree, tvb, offset - 1, 1, "TRUNCATED: %s (c=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1 );
			command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, offset - 1, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, offset - 1, 1, FALSE );
			}
		return -1;
	}

	octet1 = tvb_get_guint8( tvb, offset );

	/* seems to be an aborted MIDI-command */
	if ( octet1 & RTP_MIDI_COMMAND_STATUS_FLAG ) {
		/* this should never happen */
		if ( using_rs ) {
			command_item = proto_tree_add_text( tree, tvb, offset, 0, "ABORTED: %s (c=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1 );
			command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, rsoffset, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, rsoffset, 1, FALSE );

		} else {
			command_item = proto_tree_add_text( tree, tvb, offset - 1, 1, "ABORTED: %s (c=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1 );
			command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, offset - 1, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, offset - 1, 1, FALSE );
			}
		return -1;
	}

	/* broken: we have only one further octet */
	if ( cmd_len < 2 ) {
		if ( using_rs ) {
			command_item = proto_tree_add_text( tree, tvb, offset, 1, "TRUNCATED: %s (c=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1 );
			command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, rsoffset, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, rsoffset, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_pitch_bend_truncated, tvb, offset, 1, FALSE );

		} else {
			command_item = proto_tree_add_text( tree, tvb, offset - 1, 2, "TRUNCATED: %s (c=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1 );
			command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, offset - 1, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, offset - 1, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_pitch_bend_truncated, tvb, offset, 1, FALSE );
			}
		return -1;
	}

	octet2 = tvb_get_guint8( tvb, offset + 1 );

	/* seems to be an aborted MIDI-command */
	if ( octet2  & RTP_MIDI_COMMAND_STATUS_FLAG ) {
		if ( using_rs ) {
			command_item = proto_tree_add_text( tree, tvb, offset, 1, "ABORTED: %s (c=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1 );
			command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, rsoffset, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, rsoffset, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_pitch_bend_truncated, tvb, offset, 1, FALSE );

		} else {
			command_item = proto_tree_add_text( tree, tvb, offset - 1, 2, "ABORTED: %s (c=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1 );
			command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, offset - 1, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, offset - 1, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_pitch_bend_truncated, tvb, offset, 1, FALSE );
			}
		return -1;
	}

	pitch = ( octet1 << 7 ) | octet2;

	if ( using_rs ) {
		command_item = proto_tree_add_text( tree, tvb, offset, 2, "%s (c=%d, pb=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1, pitch );
		command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
		proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, rsoffset, 1, FALSE );
		proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, rsoffset, 1, FALSE );
		proto_tree_add_item ( command_tree, hf_rtp_midi_pitch_bend, tvb, offset, 2, FALSE );
	} else {
		command_item = proto_tree_add_text(tree, tvb, offset - 1, 3, "%s (c=%d, pb=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1, pitch );
		command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
		proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, offset - 1, 1, FALSE );
		proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, offset - 1, 1, FALSE );
		proto_tree_add_item ( command_tree, hf_rtp_midi_pitch_bend, tvb, offset, 2, FALSE );
	}

	if ( cmd_count ) {
		col_append_fstr(pinfo->cinfo, COL_INFO,", %s (c=%d, pb=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1, pitch );
	} else {
		col_append_fstr(pinfo->cinfo, COL_INFO, "%s (c=%d, pb=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1, pitch );
	}
	return 2;
}




/*
 * Here program_change is decoded.
 */
static int
decode_program_change(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned int cmd_count, unsigned int offset, unsigned int cmd_len, guint8 status, unsigned int rsoffset, gboolean using_rs ) {

	guint8		 program	= 0;
	const gchar	*status_str	= NULL;
	proto_item	*command_item	= NULL;
	proto_tree	*command_tree	= NULL;

	status_str = val_to_str( status >> 4, rtp_midi_channel_status, "" );

	/* broken: we have no further data */
	if ( !cmd_len ) {
		/* this should never happen */
		if ( using_rs ) {
			command_item = proto_tree_add_text( tree, tvb, offset, 0, "TRUNCATED: %s (c=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1 );
			command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, rsoffset, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, rsoffset, 1, FALSE );

		} else {
			command_item = proto_tree_add_text( tree, tvb, offset - 1, 1, "TRUNCATED: %s (c=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1);
			command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, offset - 1, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, offset - 1, 1, FALSE );
			}
		return -1;
	}

	program = tvb_get_guint8( tvb, offset );

	/* seems to be an aborted MIDI-command */
	if ( program  & RTP_MIDI_COMMAND_STATUS_FLAG ) {
		/* this should never happen */
		if ( using_rs ) {
			command_item = proto_tree_add_text( tree, tvb, offset, 0, "ABORTED: %s (c=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1 );
			command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, rsoffset, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, rsoffset, 1, FALSE );

		} else {
			command_item = proto_tree_add_text( tree, tvb, offset - 1, 1, "ABORTED: %s (c=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1);
			command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, offset - 1, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, offset - 1, 1, FALSE );
			}
		return -1;
	}


	if ( using_rs ) {
		command_item = proto_tree_add_text( tree, tvb, offset, 1, "%s (c=%d, p=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1, program );
		command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
		proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, rsoffset, 1, FALSE );
		proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, rsoffset, 1, FALSE );
		proto_tree_add_item ( command_tree, hf_rtp_midi_program, tvb, offset, 1, FALSE );
	} else {
		command_item = proto_tree_add_text(tree, tvb, offset - 1, 2, "%s (c=%d, p=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1, program );
		command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
		proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, offset - 1, 1, FALSE );
		proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, offset - 1, 1, FALSE );
		proto_tree_add_item ( command_tree, hf_rtp_midi_program, tvb, offset, 1, 0);
	}

	if ( cmd_count ) {
		col_append_fstr(pinfo->cinfo, COL_INFO,", %s (c=%d, p=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1, program );
	} else {
		col_append_fstr(pinfo->cinfo, COL_INFO, "%s (c=%d, p=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1, program );
	}


return 1;
}



/*
 * Here control change is decoded.
 */
static int
decode_control_change(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned int cmd_count, unsigned int offset, unsigned int cmd_len, guint8 status, unsigned int rsoffset, gboolean using_rs ) {

	guint8		 controller	= 0;
	guint8		 value		= 0;
	const gchar	*status_str	= NULL;
	const gchar	*ctrl_str	= NULL;
	proto_item	*command_item	= NULL;
	proto_tree	*command_tree	= NULL;

	status_str = val_to_str( status >> 4, rtp_midi_channel_status, "" );

	/* broken: we have no further data */
	if ( !cmd_len ) {
		/* this should never happen */
		if ( using_rs ) {
			command_item = proto_tree_add_text( tree, tvb, offset, 0, "TRUNCATED: %s (c=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1 );
			command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, rsoffset, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, rsoffset, 1, FALSE );

		} else {
			command_item = proto_tree_add_text( tree, tvb, offset - 1, 1, "TRUNCATED: %s (c=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1 );
			command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, offset - 1, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, offset - 1, 1, FALSE );
			}
		return -1;
	}

	controller = tvb_get_guint8( tvb, offset );

	/* seems to be an aborted MIDI-command */
	if ( controller & RTP_MIDI_COMMAND_STATUS_FLAG ) {
		/* this should never happen */
		if ( using_rs ) {
			command_item = proto_tree_add_text( tree, tvb, offset, 0, "ABORTED: %s (c=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1 );
			command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, rsoffset, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, rsoffset, 1, FALSE );

		} else {
			command_item = proto_tree_add_text( tree, tvb, offset - 1, 1, "ABORTED: %s (c=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1 );
			command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, offset - 1, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, offset - 1, 1, FALSE );
			}
		return -1;
	}

	ctrl_str = val_to_str( controller, rtp_midi_controller_values, "Unknown: %d" );

	/* broken: we have only one further octet */
	if ( cmd_len < 2 ) {
		if ( using_rs ) {
			command_item = proto_tree_add_text( tree, tvb, offset, 1, "TRUNCATED: %s (c=%d, ctrl=%s)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1, ctrl_str );
			command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, rsoffset, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, rsoffset, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_note, tvb, offset, 1, FALSE );

		} else {
			command_item = proto_tree_add_text( tree, tvb, offset - 1, 2, "TRUNCATED: %s (c=%d, ctrl=%s)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1, ctrl_str );
			command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, offset - 1, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, offset - 1, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_note, tvb, offset, 1, FALSE );
			}
		return -1;
	}

	value = tvb_get_guint8( tvb, offset + 1 );

	/* seems to be an aborted MIDI-command */
	if ( value  & RTP_MIDI_COMMAND_STATUS_FLAG ) {
		if ( using_rs ) {
			command_item = proto_tree_add_text( tree, tvb, offset, 1, "ABORTED: %s (c=%d, ctrl=%s)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1, ctrl_str );
			command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, rsoffset, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, rsoffset, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_controller, tvb, offset, 1, FALSE );

		} else {
			command_item = proto_tree_add_text( tree, tvb, offset - 1, 2, "ABORTED: %s (c=%d, ctrl=%s)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1, ctrl_str );
			command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, offset - 1, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, offset - 1, 1, FALSE );
			proto_tree_add_item ( command_tree, hf_rtp_midi_controller, tvb, offset, 1, FALSE );
			}
		return -1;
	}


	if ( using_rs ) {
		command_item = proto_tree_add_text( tree, tvb, offset, 2, "%s (c=%d, ctrl=%s, v=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1, ctrl_str, value );
		command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
		proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, rsoffset, 1, FALSE );
		proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, rsoffset, 1, FALSE );
		proto_tree_add_item ( command_tree, hf_rtp_midi_controller, tvb, offset, 1, FALSE );
		proto_tree_add_item ( command_tree, hf_rtp_midi_controller_value, tvb, offset + 1, 1, FALSE );
	} else {
		command_item = proto_tree_add_text(tree, tvb, offset - 1, 3, "%s (c=%d, ctrl=%s, v=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1, ctrl_str, value );
		command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
		proto_tree_add_item ( command_tree, hf_rtp_midi_channel_status, tvb, offset - 1, 1, FALSE );
		proto_tree_add_item ( command_tree, hf_rtp_midi_channel, tvb, offset - 1, 1, FALSE );
		proto_tree_add_item ( command_tree, hf_rtp_midi_controller, tvb, offset, 1, FALSE );
		proto_tree_add_item ( command_tree, hf_rtp_midi_controller_value, tvb, offset + 1, 1, 0);
	}

	if ( cmd_count ) {
		col_append_fstr(pinfo->cinfo, COL_INFO,", %s (c=%d, ctrl=%s, p=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1, ctrl_str, value );
	} else {
		col_append_fstr(pinfo->cinfo, COL_INFO, "%s (c=%d, ctrl=%s, p=%d)", status_str, ( status & RTP_MIDI_CHANNEL_MASK ) + 1, ctrl_str, value );
	}
	return 2;
}


/*
 * Here a Sysex-Common Non-Realtime Sample Dump Header command is decoded.
 */
static unsigned int
decode_sysex_common_nrt_sd_hdr( tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned int offset, unsigned int data_len ) {
	/* TODO: decode sd stuff */
	return 0;
}

/*
 * Here a Sysex-Common Non-Realtime Sample Dump Packet command is decoded.
 */
static unsigned int
decode_sysex_common_nrt_sd_packet( tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned int offset, unsigned int data_len ) {
	/* TODO: decode sd stuff */
	return 0;
}

/*
 * Here a Sysex-Common Non-Realtime Sample Dump Request command is decoded.
 */
static unsigned int
decode_sysex_common_nrt_sd_req( tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned int offset, unsigned int data_len ) {
	/* TODO: decode sd stuff */
	return 0;
}

/*
 * Here a Sysex-Common Non-Realtime MTC command is decoded.
 */
static unsigned int
decode_sysex_common_nrt_mtc( tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned int offset, unsigned int data_len ) {
	guint8		 sub_id		= 0;
	int		 consumed	= 0;

	if ( data_len < 1 )
		return consumed;

	/* first we retrieve the sub-command */
	sub_id = tvb_get_guint8( tvb, offset );
	proto_tree_add_item ( tree, hf_rtp_midi_sysex_common_nrt_mtc, tvb, offset, 1, FALSE );
	offset++;
	data_len--;
	consumed++;

	/* TODO: decode MTC stuff */
	return consumed;
}

/*
 * Here a Sysex-Common Non-Realtime Sample Dump Extension command is decoded.
 */
static unsigned int
decode_sysex_common_nrt_sd_ext( tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned int offset, unsigned int data_len ) {
	guint8		 sub_id		= 0;
	int		 consumed	= 0;

	if ( data_len < 1 )
		return consumed;

	/* first we retrieve the sub-command */
	sub_id = tvb_get_guint8( tvb, offset );
	proto_tree_add_item ( tree, hf_rtp_midi_sysex_common_nrt_sd_ext, tvb, offset, 1, FALSE );
	offset++;
	data_len--;
	consumed++;

	/* TODO: decode SD Ext stuff */
	return consumed;
}

/*
 * Here a Sysex-Common Non-Realtime General Information command is decoded.
 */
static unsigned int
decode_sysex_common_nrt_gi( tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned int offset, unsigned int data_len ) {
	guint8		 sub_id		= 0;
	guint8		 manu_short	= 0;
	int		 consumed	= 0;

	if ( data_len < 1 )
		return consumed;

	/* first we retrieve the sub-command */
	sub_id = tvb_get_guint8( tvb, offset );
	proto_tree_add_item ( tree, hf_rtp_midi_sysex_common_nrt_gi, tvb, offset, 1, FALSE );
	offset++;
	data_len--;
	consumed++;

	if ( data_len < 1 ) {
		return consumed;
	}

	/* This is the only sub-command that we know that we can decode... */
	if ( sub_id != RTP_MIDI_SYSEX_COMMON_NRT_GI_IDENTITY_REPLY ) {
		return consumed;
	}

	/* the manu-id should follow */
	proto_tree_add_item( tree, hf_rtp_midi_manu_short, tvb, offset, 1, FALSE );
	manu_short = tvb_get_guint8( tvb, offset );
	offset++;
	data_len--;
	consumed++;

	/* possibly escaped to be a long manu-id */
	if ( manu_short == RTP_MIDI_MANU_SHORT_ISLONG ) {
		if ( data_len < 2 ) {
			return consumed;
		}
		proto_tree_add_item( tree, hf_rtp_midi_manu_long, tvb, offset, 2, FALSE );
		offset +=2 ;
		data_len -= 2;
		consumed += 2;
	}

	/* lets see if we can get the device family for this unit */
	if ( data_len < 2 ) {
		return consumed;
	}
	proto_tree_add_item( tree, hf_rtp_midi_sysex_common_nrt_gi_device_family, tvb, offset, 2, FALSE );
	offset +=2 ;
	data_len -= 2;
	consumed += 2;

	/* ...also get the detailed member of this family... */
	if ( data_len < 2 ) {
		return consumed;
	}
	proto_tree_add_item( tree, hf_rtp_midi_sysex_common_nrt_gi_device_family_member, tvb, offset, 2, FALSE );
	offset +=2 ;
	data_len -= 2;
	consumed += 2;

	/* and even the software-revision-level */
	if ( data_len < 4 ) {
		return consumed;
	}
	proto_tree_add_item( tree, hf_rtp_midi_sysex_common_nrt_gi_software_rev, tvb, offset, 4, FALSE );
	offset +=4 ;
	data_len -= 4;
	consumed += 4;

	return consumed;
}

/*
 * Here a Sysex-Common Non-Realtime File Dump command is decoded.
 */
static unsigned int
decode_sysex_common_nrt_fd( tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned int offset, unsigned int data_len ) {
	guint8		 sub_id		= 0;
	int		 consumed	= 0;

	if ( data_len < 1 )
		return consumed;

	/* first we retrieve the sub-command */
	sub_id = tvb_get_guint8( tvb, offset );
	proto_tree_add_item ( tree, hf_rtp_midi_sysex_common_nrt_fd, tvb, offset, 1, FALSE );
	offset++;
	data_len--;
	consumed++;

	/* TODO: decode File Dump stuff */
	return consumed;
}

/*
 * Here a Sysex-Common Non-Realtime MIDI Tuning Standard command is decoded.
 */
static unsigned int
decode_sysex_common_nrt_tuning( tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned int offset, unsigned int data_len ) {
	guint8		 sub_id		= 0;
	int		 consumed	= 0;

	if ( data_len < 1 )
		return consumed;

	/* first we retrieve the sub-command */
	sub_id = tvb_get_guint8( tvb, offset );
	proto_tree_add_item ( tree, hf_rtp_midi_sysex_common_nrt_tuning, tvb, offset, 1, FALSE );
	offset++;
	data_len--;
	consumed++;

	/* TODO: decode MIDI Tuning stuff */
	return consumed;
}

/*
 * Here a Sysex-Common Non-Realtime General MIDI command is decoded.
 */
static unsigned int
decode_sysex_common_nrt_gm( tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned int offset, unsigned int data_len ) {
	guint8		 sub_id		= 0;
	int		 consumed	= 0;

	if ( data_len < 1 )
		return consumed;

	/* first we retrieve the sub-command */
	sub_id = tvb_get_guint8( tvb, offset );
	proto_tree_add_item ( tree, hf_rtp_midi_sysex_common_nrt_gm, tvb, offset, 1, FALSE );
	offset++;
	data_len--;
	consumed++;

	return consumed;
}

/*
 * Here a Sysex-Common Non-Realtime Downloadable Sounds command is decoded.
 */
static unsigned int
decode_sysex_common_nrt_dls( tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned int offset, unsigned int data_len ) {
	guint8		 sub_id		= 0;
	int		 consumed	= 0;

	if ( data_len < 1 )
		return consumed;

	/* first we retrieve the sub-command */
	sub_id = tvb_get_guint8( tvb, offset );
	proto_tree_add_item ( tree, hf_rtp_midi_sysex_common_nrt_dls, tvb, offset, 1, FALSE );
	offset++;
	data_len--;
	consumed++;

	/* TODO: decode DLS stuff */
	return consumed;
}

/*
 * Here a Sysex-Common Non-Realtime End Of File command is decoded.
 */
static unsigned int
decode_sysex_common_nrt_eof( tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned int offset, unsigned int data_len ) {
	int		 consumed	= 0;

	if ( data_len < 1 )
		return consumed;

	/* we only have a packet-number */
	proto_tree_add_item ( tree, hf_rtp_midi_sysex_common_nrt_sd_packet_number, tvb, offset, 1, FALSE );
	offset++;
	data_len--;
	consumed++;

	return consumed;
}

/*
 * Here a Sysex-Common Non-Realtime Wait command is decoded.
 */
static unsigned int
decode_sysex_common_nrt_wait( tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned int offset, unsigned int data_len ) {
	int		 consumed	= 0;

	if ( data_len < 1 )
		return consumed;

	/* we only have a packet-number */
	proto_tree_add_item ( tree, hf_rtp_midi_sysex_common_nrt_sd_packet_number, tvb, offset, 1, FALSE );
	offset++;
	data_len--;
	consumed++;

	return consumed;
}

/*
 * Here a Sysex-Common Non-Realtime Cancel command is decoded.
 */
static unsigned int
decode_sysex_common_nrt_cancel( tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned int offset, unsigned int data_len ) {
	int		 consumed	= 0;

	if ( data_len < 1 )
		return consumed;

	/* we only have a packet-number */
	proto_tree_add_item ( tree, hf_rtp_midi_sysex_common_nrt_sd_packet_number, tvb, offset, 1, FALSE );
	offset++;
	data_len--;
	consumed++;

	return consumed;
}

/*
 * Here a Sysex-Common Non-Realtime NAK command is decoded.
 */
static unsigned int
decode_sysex_common_nrt_nak( tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned int offset, unsigned int data_len ) {
	int		 consumed	= 0;

	if ( data_len < 1 )
		return consumed;

	/* we only have a packet-number */
	proto_tree_add_item ( tree, hf_rtp_midi_sysex_common_nrt_sd_packet_number, tvb, offset, 1, FALSE );
	offset++;
	data_len--;
	consumed++;

	return consumed;
}

/*
 * Here a Sysex-Common Non-Realtime ACK command is decoded.
 */
static unsigned int
decode_sysex_common_nrt_ack( tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned int offset, unsigned int data_len ) {
	int		 consumed	= 0;

	if ( data_len < 1 )
		return consumed;

	/* we only have a packet-number */
	proto_tree_add_item ( tree, hf_rtp_midi_sysex_common_nrt_sd_packet_number, tvb, offset, 1, FALSE );
	offset++;
	data_len--;
	consumed++;

	return consumed;
}



/*
 * Here a Sysex-Common Non-Realtime command is decoded.
 */
static unsigned int
decode_sysex_common_nrt( tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned int offset, unsigned int data_len ) {
	guint8		 common_nrt	= 0;
	guint8		 device_id	= 0;
	const gchar	*nrt_str	= NULL;
	proto_item	*command_item	= NULL;
	proto_tree	*command_tree	= NULL;
	int		 consumed	= 0;
	int		 ext_consumed	= 0;

	if ( data_len < 1 )
		return consumed;

	device_id = tvb_get_guint8( tvb, offset );
	proto_tree_add_item ( tree, hf_rtp_midi_sysex_common_device_id, tvb, offset, 1, FALSE );
	offset++;
	data_len--;
	consumed++;

	if ( data_len < 1 ) {
		return consumed;
	}

	common_nrt = tvb_get_guint8( tvb, offset );

	nrt_str = val_to_str( common_nrt, rtp_midi_sysex_common_nrt, "Unknown 0x%02x" );

	command_item = proto_tree_add_text(tree, tvb, offset, data_len, nrt_str );
	command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_sysex_common_nrt );
	proto_tree_add_item ( command_tree, hf_rtp_midi_sysex_common_non_realtime, tvb, offset, 1, FALSE );

	offset++;
	data_len--;
	consumed++;

	switch ( common_nrt ) {
		case RTP_MIDI_SYSEX_COMMON_NRT_SAMPLE_DUMP_HEADER:
			ext_consumed = decode_sysex_common_nrt_sd_hdr( tvb, pinfo, command_tree, offset, data_len );
			break;
		case RTP_MIDI_SYSEX_COMMON_NRT_SAMPLE_DATA_PACKET:
			ext_consumed = decode_sysex_common_nrt_sd_packet( tvb, pinfo, command_tree, offset, data_len );
			break;
		case RTP_MIDI_SYSEX_COMMON_NRT_SAMPLE_DUMP_REQUEST:
			ext_consumed = decode_sysex_common_nrt_sd_req( tvb, pinfo, command_tree, offset, data_len );
			break;
		case RTP_MIDI_SYSEX_COMMON_NRT_MTC:
			ext_consumed = decode_sysex_common_nrt_mtc( tvb, pinfo, command_tree, offset, data_len );
			break;
		case RTP_MIDI_SYSEX_COMMON_NRT_SAMPLE_DUMP_EXTENSIONS:
			ext_consumed = decode_sysex_common_nrt_sd_ext( tvb, pinfo, command_tree, offset, data_len );
			break;
		case RTP_MIDI_SYSEX_COMMON_NRT_GENERAL_INFORMATION:
			ext_consumed = decode_sysex_common_nrt_gi( tvb, pinfo, command_tree, offset, data_len );
			break;
		case RTP_MIDI_SYSEX_COMMON_NRT_FILE_DUMP:
			ext_consumed = decode_sysex_common_nrt_fd( tvb, pinfo, command_tree, offset, data_len );
			break;
		case RTP_MIDI_SYSEX_COMMON_NRT_MIDI_TUNING_STANDARD:
			ext_consumed = decode_sysex_common_nrt_tuning( tvb, pinfo, command_tree, offset, data_len );
			break;
		case RTP_MIDI_SYSEX_COMMON_NRT_GENERAL_MIDI:
			ext_consumed = decode_sysex_common_nrt_gm( tvb, pinfo, command_tree, offset, data_len );
			break;
		case RTP_MIDI_SYSEX_COMMON_NRT_DOWNLOADABLE_SOUNDS:		/* ??? not in Spec */
			ext_consumed = decode_sysex_common_nrt_dls( tvb, pinfo, command_tree, offset, data_len );
			break;
		case RTP_MIDI_SYSEX_COMMON_NRT_END_OF_FILE:
			ext_consumed = decode_sysex_common_nrt_eof( tvb, pinfo, command_tree, offset, data_len );
			break;
		case RTP_MIDI_SYSEX_COMMON_NRT_WAIT:
			ext_consumed = decode_sysex_common_nrt_wait( tvb, pinfo, command_tree, offset, data_len );
			break;
		case RTP_MIDI_SYSEX_COMMON_NRT_CANCEL:
			ext_consumed = decode_sysex_common_nrt_cancel( tvb, pinfo, command_tree, offset, data_len );
			break;
		case RTP_MIDI_SYSEX_COMMON_NRT_NAK:
			ext_consumed = decode_sysex_common_nrt_nak( tvb, pinfo, command_tree, offset, data_len );
			break;
		case RTP_MIDI_SYSEX_COMMON_NRT_ACK:
			ext_consumed = decode_sysex_common_nrt_ack( tvb, pinfo, command_tree, offset, data_len );
			break;
	}

	if ( ext_consumed < 0 ) {
		return -1;
	}

	/* set our pointers correct to move past already decoded data */
	offset += ext_consumed;
	data_len -= ext_consumed;
	consumed += ext_consumed;

	if ( data_len > 0 ) {
		proto_tree_add_item( command_tree, hf_rtp_midi_unknown_data, tvb, offset, data_len, FALSE );
		consumed += data_len;
	}

	return consumed;
}






/*
 * Here a Sysex-Common Realtime MIDI Time Code command is decoded.
 */
static unsigned int
decode_sysex_common_rt_mtc( tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned int offset, unsigned int data_len ) {
	guint8		 sub_id		= 0;
	int		 consumed	= 0;

	if ( data_len < 1 )
		return consumed;

	/* first we retrieve the sub-command */
	sub_id = tvb_get_guint8( tvb, offset );
	proto_tree_add_item ( tree, hf_rtp_midi_sysex_common_rt_mtc, tvb, offset, 1, FALSE );
	offset++;
	data_len--;
	consumed++;

	/* TODO: decode MTC stuff */
	return consumed;
}

/*
 * Here a Sysex-Common Realtime MIDI Show Control command is decoded.
 */
static unsigned int
decode_sysex_common_rt_sc( tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned int offset, unsigned int data_len ) {
	guint8		 sub_id		= 0;
	int		 consumed	= 0;

	if ( data_len < 1 )
		return consumed;

	/* first we retrieve the sub-command */
	sub_id = tvb_get_guint8( tvb, offset );
	proto_tree_add_item ( tree, hf_rtp_midi_sysex_common_rt_sc, tvb, offset, 1, FALSE );
	offset++;
	data_len--;
	consumed++;

	/* TODO: decode Show Control stuff */
	return consumed;
}

/*
 * Here a Sysex-Common Realtime Notation Information command is decoded.
 */
static unsigned int
decode_sysex_common_rt_ni( tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned int offset, unsigned int data_len ) {
	guint8		 sub_id		= 0;
	int		 consumed	= 0;

	if ( data_len < 1 )
		return consumed;

	/* first we retrieve the sub-command */
	sub_id = tvb_get_guint8( tvb, offset );
	proto_tree_add_item ( tree, hf_rtp_midi_sysex_common_rt_ni, tvb, offset, 1, FALSE );
	offset++;
	data_len--;
	consumed++;

	/* TODO: decode Notation Information stuff */
	return consumed;
}

/*
 * Here a Sysex-Common Realtime Device Control command is decoded.
 */
static unsigned int
decode_sysex_common_rt_dc( tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned int offset, unsigned int data_len ) {
	guint8		 sub_id		= 0;
	int		 consumed	= 0;

	if ( data_len < 1 )
		return consumed;

	/* first we retrieve the sub-command */
	sub_id = tvb_get_guint8( tvb, offset );
	proto_tree_add_item ( tree, hf_rtp_midi_sysex_common_rt_dc, tvb, offset, 1, FALSE );
	offset++;
	data_len--;
	consumed++;

	/* TODO: decode Device Control stuff */
	return consumed;
}

/*
 * Here a Sysex-Common Realtime MIDI Time Code Cueing command is decoded.
 */
static unsigned int
decode_sysex_common_rt_mtc_cue( tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned int offset, unsigned int data_len ) {
	guint8		 sub_id		= 0;
	int		 consumed	= 0;

	if ( data_len < 1 )
		return consumed;

	/* first we retrieve the sub-command */
	sub_id = tvb_get_guint8( tvb, offset );
	proto_tree_add_item ( tree, hf_rtp_midi_sysex_common_rt_mtc_cueing, tvb, offset, 1, FALSE );
	offset++;
	data_len--;
	consumed++;

	/* TODO: decode MTC cue stuff */
	return consumed;
}

/*
 * Here a Sysex-Common Realtime MIDI Machine Control command is decoded.
 */
static unsigned int
decode_sysex_common_rt_mmc_command( tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned int offset, unsigned int data_len ) {
	guint8		 sub_id		= 0;
	int		 consumed	= 0;

	if ( data_len < 1 )
		return consumed;

	/* first we retrieve the sub-command */
	sub_id = tvb_get_guint8( tvb, offset );
	proto_tree_add_item ( tree, hf_rtp_midi_sysex_common_rt_mmc_commands, tvb, offset, 1, FALSE );
	offset++;
	data_len--;
	consumed++;

	/* TODO: decode MMC stuff */
	return consumed;
}

/*
 * Here a Sysex-Common Realtime MIDI Machine Control response is decoded.
 */
static unsigned int
decode_sysex_common_rt_mmc_response( tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned int offset, unsigned int data_len ) {
	guint8		 sub_id		= 0;
	int		 consumed	= 0;

	if ( data_len < 1 )
		return consumed;

	/* first we retrieve the sub-command */
	sub_id = tvb_get_guint8( tvb, offset );
	proto_tree_add_item ( tree, hf_rtp_midi_sysex_common_rt_mmc_responses, tvb, offset, 1, FALSE );
	offset++;
	data_len--;
	consumed++;

	/* TODO: decode MMC stuff */
	return consumed;
}

/*
 * Here a Sysex-Common Realtime MIDI Tuning Standard command is decoded.
 */
static unsigned int
decode_sysex_common_rt_tuning( tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned int offset, unsigned int data_len ) {
	guint8		 sub_id		= 0;
	int		 consumed	= 0;

	if ( data_len < 1 )
		return consumed;

	/* first we retrieve the sub-command */
	sub_id = tvb_get_guint8( tvb, offset );
	proto_tree_add_item ( tree, hf_rtp_midi_sysex_common_rt_tuning, tvb, offset, 1, FALSE );
	offset++;
	data_len--;
	consumed++;

	/* TODO: decode Tuning stuff */
	return consumed;
}


/*
 * Here a Sysex-Common Realtime command is decoded.
 */
static unsigned int
decode_sysex_common_rt( tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned int offset, unsigned int data_len ) {
	guint8		 common_rt	= 0;
	guint8		 device_id	= 0;
	const gchar	*rt_str		= NULL;
	proto_item	*command_item	= NULL;
	proto_tree	*command_tree	= NULL;
	int		 consumed	= 0;
	int		 ext_consumed	= 0;

	if ( data_len < 1 )
		return consumed;

	device_id = tvb_get_guint8( tvb, offset );
	proto_tree_add_item ( tree, hf_rtp_midi_sysex_common_device_id, tvb, offset, 1, FALSE );
	offset++;
	data_len--;
	consumed++;

	if ( data_len < 1 ) {
		return consumed;
	}

	common_rt = tvb_get_guint8( tvb, offset );
	rt_str = val_to_str( common_rt, rtp_midi_sysex_common_rt, "Unknown 0x%02x" );

	command_item = proto_tree_add_text(tree, tvb, offset, data_len, rt_str );
	command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_sysex_common_rt );
	proto_tree_add_item ( command_tree, hf_rtp_midi_sysex_common_realtime, tvb, offset, 1, FALSE );

	data_len--;
	offset++;
	consumed++;

	switch ( common_rt ) {
		case RTP_MIDI_SYSEX_COMMON_RT_MIDI_TIME_CODE:
			ext_consumed = decode_sysex_common_rt_mtc( tvb, pinfo, command_tree, offset, data_len );
			break;
		case RTP_MIDI_SYSEX_COMMON_RT_MIDI_SHOW_CONTROL:
			ext_consumed = decode_sysex_common_rt_sc( tvb, pinfo, command_tree, offset, data_len );
			break;
		case RTP_MIDI_SYSEX_COMMON_RT_NOTATION_INFORMATION:
			ext_consumed = decode_sysex_common_rt_ni( tvb, pinfo, command_tree, offset, data_len );
			break;
		case RTP_MIDI_SYSEX_COMMON_RT_DEVICE_CONTROL:
			ext_consumed = decode_sysex_common_rt_dc( tvb, pinfo, command_tree, offset, data_len );
			break;
		case RTP_MIDI_SYSEX_COMMON_RT_MTC_CUEING:
			ext_consumed = decode_sysex_common_rt_mtc_cue( tvb, pinfo, command_tree, offset, data_len );
			break;
		case RTP_MIDI_SYSEX_COMMON_RT_MIDI_MACHINE_CONTROL_COMMAND:
			ext_consumed = decode_sysex_common_rt_mmc_command( tvb, pinfo, command_tree, offset, data_len );
			break;
		case RTP_MIDI_SYSEX_COMMON_RT_MIDI_MACHINE_CONTROL_RESPONSE:
			ext_consumed = decode_sysex_common_rt_mmc_response( tvb, pinfo, command_tree, offset, data_len );
			break;
		case RTP_MIDI_SYSEX_COMMON_RT_MIDI_TUNING_STANDARD:
			ext_consumed = decode_sysex_common_rt_tuning( tvb, pinfo, command_tree, offset, data_len );
			break;
	}

	if ( ext_consumed < 0 ) {
		return -1;
	}

	/* set our pointers correct to move past already decoded data */
	offset += ext_consumed;
	data_len -= ext_consumed;
	consumed += ext_consumed;

	if ( data_len > 0 ) {
		proto_tree_add_item( command_tree, hf_rtp_midi_unknown_data, tvb, offset, data_len, FALSE );
		consumed += data_len;
	}

	return consumed;
}



/*
 * Here a Sysex-Start command is decoded.
 */
static unsigned int
decode_sysex_start(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned int cmd_count, unsigned int offset, unsigned int cmd_len ) {
	guint8		 octet		= 0;
	guint8		 manu_short	= 0;
	const gchar	*status_str	= NULL;
	proto_item	*command_item	= NULL;
	proto_tree	*command_tree	= NULL;
	int		 consumed	= 0;
	int		 data_len	= 0;
	int		 ext_consumed	= 0;

	status_str = val_to_str( RTP_MIDI_STATUS_COMMON_SYSEX_END, rtp_midi_common_status, "" );

	/* we need to parse "away" data until the next command */
	while ( cmd_len ) {
		octet = tvb_get_guint8( tvb, offset + consumed );
		consumed++;

		/*
		 * lets check if we have a "normal" end -
		 * if so, the last status-byte is consumed
		 * as it belongs to the message
		 */
		if ( octet == RTP_MIDI_STATUS_COMMON_SYSEX_END ) {
			status_str = rtp_midi_common_status_sysex_segment_complete;
		} else if ( octet == RTP_MIDI_STATUS_COMMON_SYSEX_START ) {
			status_str = rtp_midi_common_status_sysex_segment_start;
		} else if ( octet == RTP_MIDI_STATUS_COMMON_UNDEFINED_F4 ) {
			status_str = rtp_midi_common_status_sysex_cancel;
		}

		/* Is this command through? */
		if ( octet & RTP_MIDI_COMMAND_STATUS_FLAG ) {
			break;
		}
	}

	/* the unknown data we have does not include the trailing status-command */
	data_len = consumed - 1;

	command_item = proto_tree_add_text(tree, tvb, offset - 1, 1 + consumed, status_str );
	command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
	proto_tree_add_item ( command_tree, hf_rtp_midi_common_status, tvb, offset - 1, 1, FALSE );

	/* lets see if we have enough data for the sysex-id */
	/* TODO: Check if this len-check is correct - and also the one for the long-manu-id below */
	if ( data_len > 1 ) {
		proto_tree_add_item( command_tree, hf_rtp_midi_manu_short, tvb, offset, 1, FALSE );
		manu_short = tvb_get_guint8( tvb, offset );
		offset++;
		data_len--;

		switch( manu_short ) {
			case RTP_MIDI_MANU_SHORT_NON_REALTIME_UNIVERSAL:
				ext_consumed = decode_sysex_common_nrt( tvb, pinfo, command_tree, offset, data_len );
				break;
			case RTP_MIDI_MANU_SHORT_REALTIME_UNIVERSAL:
				ext_consumed = decode_sysex_common_rt( tvb, pinfo, command_tree, offset, data_len );
				break;
			case RTP_MIDI_MANU_SHORT_ISLONG:
				/* lets see if we have enough data to decode a long sysex-id and if we have a long id at all */
				if ( data_len > 2 ) {
					proto_tree_add_item( command_tree, hf_rtp_midi_manu_long, tvb, offset, 2, FALSE );
					offset +=2 ;
					data_len -= 2;
				}
				break;
		}
	}

	if ( ext_consumed < 0 ) {
		return -1;
	}

	/* set our pointers correct to move past already decoded data */
	offset += ext_consumed;
	data_len -= ext_consumed;

	/* just add the remaining unknown data to the tree, in case anything is still left! */
	if ( data_len ) {
		proto_tree_add_item( command_tree, hf_rtp_midi_unknown_data, tvb, offset, data_len, FALSE );
	}

	proto_tree_add_item ( command_tree, hf_rtp_midi_common_status, tvb, offset + data_len, 1, FALSE );

	if ( cmd_count ) {
		col_append_fstr(pinfo->cinfo, COL_INFO,", %s", status_str );
	} else {
		col_append_str(pinfo->cinfo, COL_INFO, status_str );
	}

	return consumed;
}

/*
 * Here the MIDI-Time-Code (MTC) Quarter Frame command is decoded.
 */
static int
decode_mtc_quarter_frame(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned int cmd_count, unsigned int offset, unsigned int cmd_len ) {
	guint8		 value		= 0;
	const gchar	*status_str	= NULL;
	proto_item	*command_item	= NULL;
	proto_tree	*command_tree	= NULL;

	status_str = val_to_str( RTP_MIDI_STATUS_COMMON_MTC_QUARTER_FRAME, rtp_midi_common_status, "" );

	/* broken: we have no further data */
	if ( !cmd_len ) {
		command_item = proto_tree_add_text( tree, tvb, offset - 1, 1, "TRUNCATED: %s ", status_str );
		command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
		proto_tree_add_item ( command_tree, hf_rtp_midi_common_status, tvb, offset - 1, 1, FALSE );
		return -1;
	}

	value = tvb_get_guint8( tvb, offset );

	/* seems to be an aborted MIDI-command */
	if ( value & RTP_MIDI_COMMAND_STATUS_FLAG ) {
		command_item = proto_tree_add_text( tree, tvb, offset - 1, 1, "ABORTED: %s", status_str );
		command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
		proto_tree_add_item ( command_tree, hf_rtp_midi_common_status, tvb, offset - 1, 1, FALSE );
		return -1;
	}


	command_item = proto_tree_add_text(tree, tvb, offset - 1, 2, status_str );
	command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
	proto_tree_add_item ( command_tree, hf_rtp_midi_common_status, tvb, offset - 1, 1, FALSE );
	proto_tree_add_item ( command_tree, hf_rtp_midi_quarter_frame_type, tvb, offset, 1, FALSE );
	proto_tree_add_item ( command_tree, hf_rtp_midi_quarter_frame_value, tvb, offset + 1, 1, FALSE );

	if ( cmd_count ) {
		col_append_fstr(pinfo->cinfo, COL_INFO, ", %s", status_str );
	} else {
		col_append_str(pinfo->cinfo, COL_INFO, status_str );
	}
	return 1;
}

/*
 * Here the Song Position Pointer command is decoded.
 */
static int
decode_song_position_pointer(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned int cmd_count, unsigned int offset, unsigned int cmd_len ) {
	guint8		 octet1		= 0;
	guint8		 octet2		= 0;
	guint8		 position	= 0;
	const gchar	*status_str	= NULL;
	proto_item	*command_item	= NULL;
	proto_tree	*command_tree	= NULL;

	status_str = val_to_str( RTP_MIDI_STATUS_COMMON_SONG_POSITION_POINTER, rtp_midi_common_status, "" );

	/* broken: we have no further data */
	if ( !cmd_len ) {
		command_item = proto_tree_add_text( tree, tvb, offset - 1, 1, "TRUNCATED: %s", status_str );
		command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
		proto_tree_add_item ( command_tree, hf_rtp_midi_common_status, tvb, offset - 1, 1, FALSE );
		return -1;
	}

	octet1 = tvb_get_guint8( tvb, offset );

	/* seems to be an aborted MIDI-command */
	if ( octet1 & RTP_MIDI_COMMAND_STATUS_FLAG ) {
		command_item = proto_tree_add_text( tree, tvb, offset - 1, 1, "ABORTED: %s", status_str );
		command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
		proto_tree_add_item ( command_tree, hf_rtp_midi_common_status, tvb, offset - 1, 1, FALSE );
		return -1;
	}

	/* broken: we have only one further octet */
	if ( cmd_len < 2 ) {
		command_item = proto_tree_add_text( tree, tvb, offset - 1, 2, "TRUNCATED: %s" );
		command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
		proto_tree_add_item ( command_tree, hf_rtp_midi_common_status, tvb, offset - 1, 1, FALSE );
		proto_tree_add_item ( command_tree, hf_rtp_midi_spp_truncated, tvb, offset, 1, FALSE );
		return -1;
	}

	octet2 = tvb_get_guint8( tvb, offset + 1 );

	/* seems to be an aborted MIDI-command */
	if ( octet2  & RTP_MIDI_COMMAND_STATUS_FLAG ) {
		command_item = proto_tree_add_text( tree, tvb, offset - 1, 2, "ABORTED: %s", status_str );
		command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
		proto_tree_add_item ( command_tree, hf_rtp_midi_common_status, tvb, offset - 1, 1, FALSE );
		proto_tree_add_item ( command_tree, hf_rtp_midi_spp_truncated, tvb, offset, 1, FALSE );
		return -1;
	}

	position = ( octet1 << 7 ) | octet2;

	command_item = proto_tree_add_text(tree, tvb, offset - 1, 3, "%s (p=%d)", status_str, position );
	command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
	proto_tree_add_item ( command_tree, hf_rtp_midi_common_status, tvb, offset - 1, 1, FALSE );
	proto_tree_add_item ( command_tree, hf_rtp_midi_spp, tvb, offset, 2, FALSE );

	if ( cmd_count ) {
		col_append_fstr(pinfo->cinfo, COL_INFO,", %s (p=%d)", status_str, position );
	} else {
		col_append_fstr(pinfo->cinfo, COL_INFO, "%s (p=%d)", status_str, position );
	}
	return 2;
}

/*
 * Here a Song-Select command is decoded.
 */
static int
decode_song_select(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned int cmd_count, unsigned int offset, unsigned int cmd_len ) {
	guint8		 song_nr	= 0;
	const gchar	*status_str	= NULL;
	proto_item	*command_item	= NULL;
	proto_tree	*command_tree	= NULL;

	status_str = val_to_str( RTP_MIDI_STATUS_COMMON_SONG_SELECT, rtp_midi_common_status, "" );

	/* broken: we have no further data */
	if ( !cmd_len ) {
		command_item = proto_tree_add_text( tree, tvb, offset - 1, 1, "TRUNCATED: %s", status_str );
		command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
		proto_tree_add_item ( command_tree, hf_rtp_midi_common_status, tvb, offset - 1, 1, FALSE );
		return -1;
	}

	song_nr = tvb_get_guint8( tvb, offset );

	/* seems to be an aborted MIDI-command */
	if ( song_nr & RTP_MIDI_COMMAND_STATUS_FLAG ) {
		command_item = proto_tree_add_text( tree, tvb, offset - 1, 1, "ABORTED: %s", status_str );
		command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
		proto_tree_add_item ( command_tree, hf_rtp_midi_common_status, tvb, offset - 1, 1, FALSE );
		return -1;
	}

	command_item = proto_tree_add_text(tree, tvb, offset - 1, 2, "%s (s=%d)", status_str, song_nr );
	command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
	proto_tree_add_item ( command_tree, hf_rtp_midi_common_status, tvb, offset - 1, 1, FALSE );
	proto_tree_add_item ( command_tree, hf_rtp_midi_song_select, tvb, offset, 1, FALSE );

	if ( cmd_count ) {
		col_append_fstr(pinfo->cinfo, COL_INFO,", %s (s=%d)", status_str, song_nr );
	} else {
		col_append_fstr(pinfo->cinfo, COL_INFO, "%s (s=%d)", status_str, song_nr );
	}
	return 1;
}

/*
 * Here the undefined common-command 0xf4 is decoded.
 */
static int
decode_undefined_f4(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned int cmd_count, unsigned int offset, unsigned int cmd_len ) {
	guint8		 octet		= 0;
	const gchar	*status_str	= NULL;
	proto_item	*command_item	= NULL;
	proto_tree	*command_tree	= NULL;
	int		 consumed	= 0;

	status_str = val_to_str( RTP_MIDI_STATUS_COMMON_UNDEFINED_F4, rtp_midi_common_status, "" );

	/* we need to parse "away" data until the next command */
	while ( cmd_len ) {
		octet = tvb_get_guint8( tvb, offset + consumed );
		consumed++;

		/* Is this command through? */
		if ( octet & RTP_MIDI_COMMAND_STATUS_FLAG ) {
			break;
		}
	}

	command_item = proto_tree_add_text(tree, tvb, offset - 1, 1 + consumed, status_str );
	command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
	proto_tree_add_item ( command_tree, hf_rtp_midi_common_status, tvb, offset - 1, 1, FALSE );
	/* just add the unknown data to the tree */
	proto_tree_add_item ( command_tree, hf_rtp_midi_unknown_data, tvb, offset, consumed, FALSE );
	proto_tree_add_item ( command_tree, hf_rtp_midi_common_status, tvb, offset + consumed - 1, 1, FALSE );

	if ( cmd_count ) {
		col_append_fstr(pinfo->cinfo, COL_INFO,", %s", status_str );
	} else {
		col_append_str(pinfo->cinfo, COL_INFO, status_str );
	}
	return consumed;
}

/*
 * Here the undefined common-command 0xf5 is decoded.
 */
static int
decode_undefined_f5(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned int cmd_count, unsigned int offset, unsigned int cmd_len ) {
	guint8		 octet		= 0;
	const gchar	*status_str	= NULL;
	proto_item	*command_item	= NULL;
	proto_tree	*command_tree	= NULL;
	int		 consumed	= 0;

	status_str = val_to_str( RTP_MIDI_STATUS_COMMON_UNDEFINED_F5, rtp_midi_common_status, "" );

	/* we need to parse "away" data until the next command */
	while ( cmd_len ) {
		octet = tvb_get_guint8( tvb, offset + consumed );
		consumed++;
		/* Is this command done? */
		if ( octet & RTP_MIDI_COMMAND_STATUS_FLAG ) {
			break;
		}
	}

	command_item = proto_tree_add_text(tree, tvb, offset - 1, 1 + consumed, status_str );
	command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
	proto_tree_add_item ( command_tree, hf_rtp_midi_common_status, tvb, offset - 1, 1, FALSE );
	/* just add the unknown data to the tree */
	proto_tree_add_item ( command_tree, hf_rtp_midi_unknown_data, tvb, offset, consumed, FALSE );
	proto_tree_add_item ( command_tree, hf_rtp_midi_common_status, tvb, offset + consumed - 1, 1, FALSE );

	if ( cmd_count ) {
		col_append_fstr(pinfo->cinfo, COL_INFO,", %s", status_str );
	} else {
		col_append_str(pinfo->cinfo, COL_INFO, status_str );
	}
	return consumed;
}

/*
 * Here a Tune-Request command is decoded.
 */
static int
decode_tune_request(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned int cmd_count, unsigned int offset, unsigned int cmd_len ) {
	const gchar	*status_str	= NULL;
	proto_item	*command_item	= NULL;
	proto_tree	*command_tree	= NULL;

	status_str = val_to_str( RTP_MIDI_STATUS_COMMON_TUNE_REQUEST, rtp_midi_common_status, "" );
	command_item = proto_tree_add_text(tree, tvb, offset - 1, 1, status_str );
	command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
	proto_tree_add_item ( command_tree, hf_rtp_midi_common_status, tvb, offset - 1, 1, FALSE );

	if ( cmd_count ) {
		col_append_fstr(pinfo->cinfo, COL_INFO,", %s)", status_str );
	} else {
		col_append_str(pinfo->cinfo, COL_INFO, status_str );
	}
	return 0;
}

/*
 * Here a Sysex-End command is decoded - in RTP-MIDI this has a special semantic, it either starts a segmented Sysex-frame or a Sysex-Cancel
 */
static int
decode_sysex_end(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned int cmd_count, unsigned int offset, unsigned int cmd_len ) {
	guint8		 octet		= 0;
	const gchar	*status_str	= NULL;
	proto_item	*command_item	= NULL;
	proto_tree	*command_tree	= NULL;
	int		 consumed	= 0;

	status_str = val_to_str( RTP_MIDI_STATUS_COMMON_SYSEX_END, rtp_midi_common_status, "" );

	/* we need to parse "away" data until the next command */
	while ( cmd_len ) {
		octet = tvb_get_guint8( tvb, offset + consumed );
		consumed++;

		/*
		 * lets check if we have a "normal" end -
		 * if so, the last status-byte is consumed
		 * as it belongs to the message
		 */
		if ( octet == RTP_MIDI_STATUS_COMMON_SYSEX_END ) {
			status_str = rtp_midi_common_status_sysex_segment_end;
		} else if ( octet == RTP_MIDI_STATUS_COMMON_SYSEX_START ) {
			status_str = rtp_midi_common_status_sysex_segment;
		} else if ( octet == RTP_MIDI_STATUS_COMMON_UNDEFINED_F4 ) {
			status_str = rtp_midi_common_status_sysex_cancel;
		}

		/* Is this command through? */
		if ( octet & RTP_MIDI_COMMAND_STATUS_FLAG ) {
			break;
		}
	}

	command_item = proto_tree_add_text(tree, tvb, offset - 1, 1 + consumed, status_str );
	command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command );
	proto_tree_add_item ( command_tree, hf_rtp_midi_common_status, tvb, offset - 1, 1, FALSE );
	proto_tree_add_item ( command_tree, hf_rtp_midi_unknown_data, tvb, offset, consumed - 1, FALSE );
	proto_tree_add_item ( command_tree, hf_rtp_midi_common_status, tvb, offset + consumed - 1, 1, FALSE );


	if ( cmd_count ) {
		col_append_fstr(pinfo->cinfo, COL_INFO,", %s", status_str );
	} else {
		col_append_str(pinfo->cinfo, COL_INFO, status_str );
	}

	return consumed;
}



/*
 * Here each single MIDI-command is decoded.
 * The Status-octet is read and then the decoding
 * of the individual MIDI-commands is punted to
 * external decoders.
 */

static int
decodemidi(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned int cmd_count, unsigned int offset, unsigned int cmd_len, guint8 *runningstatus, unsigned int *rsoffset )
{
	guint8		 octet		= 0;
	const gchar	*valstr		= NULL;
	const gchar	*valstr2	= NULL;
	int		 consumed	= 0;
	int		 ext_consumed	= 0;
	gboolean	 using_rs	= FALSE;
	proto_item	*command_item     = NULL;
	proto_tree	*command_tree     = NULL;


/*	guint8		octet2;
	guint8		octet3;
	guint8		cmd;
	guint8		channel;
	proto_tree	*sysexdata_tree   = NULL;
	gboolean	foundend;
	guint16		wordvalue;
	char		helptext[200];
	char		entry[200];
	char		segtype[200];
	guint8		manu1;
	guint16		manu2;
	guint8		subid1;
	guint8		subid2;
	guint8		sysexchan;
	unsigned int	helpoffset;*/


	/* extra sanity check */
	if ( !cmd_len ) {
		return -1;
	}


	octet = tvb_get_guint8( tvb, offset );


	/* midi realtime-data -> one octet  -- unlike serial-wired MIDI realtime-commands in RTP-MIDI will
	 * not be intermingled with other MIDI-commands, so we handle this case right here and return */
	if ( octet >= 0xf8 ) {
		valstr =  val_to_str( octet, rtp_midi_common_status, "");
		command_item = proto_tree_add_text(tree, tvb, offset, 1, valstr );
		command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
		proto_tree_add_item ( command_tree, hf_rtp_midi_common_status, tvb, offset, 1, FALSE );

		if ( cmd_count ) {
			col_add_fstr(pinfo->cinfo, COL_INFO,", %s", valstr );
		} else {
			col_add_str(pinfo->cinfo, COL_INFO, valstr );
		}
		return 1;
	}



	/* see if this first octet is a status message */
	if ( ( octet & RTP_MIDI_COMMAND_STATUS_FLAG ) == 0 ) {
		/* if we have no running status yet -> error */
		if ( ( ( *runningstatus ) & RTP_MIDI_COMMAND_STATUS_FLAG ) == 0 ) {
			return -1;
		}
		/* our first octet is "virtual" coming from a preceding MIDI-command,
		 * so actually we have not really consumed anything yet */
		octet = *runningstatus;
		using_rs = TRUE;
	} else {
		/* We have a "real" status-byte */
		using_rs = FALSE;

		/* Let's see how this octet influences our running-status */
		/* if we have a "normal" MIDI-command then the new status replaces the current running-status */
		if ( octet < 0xf0 ) {
			*rsoffset = offset ;
			*runningstatus = octet;
		} else {
			/* system-realtime-commands maintain the current running-status
			 * other system-commands clear the running-status, since we
			 * already handled realtime, we can reset it here */
			*runningstatus = 0;
		}

		/* lets update our pointers */
		consumed++;
		cmd_len--;
		offset++;
	}




	/* non-system MIDI-commands encode the command in the high nibble and the channel
	 * in the low nibble - so we will take care of those cases next */

	if ( octet < 0xf0 ) {
		switch ( octet >> 4 ) {
			case RTP_MIDI_STATUS_CHANNEL_NOTE_OFF:
				ext_consumed = decode_note_off( tvb, pinfo, tree, cmd_count, offset,  cmd_len, octet, *rsoffset, using_rs );
				break;
			case RTP_MIDI_STATUS_CHANNEL_NOTE_ON:
				ext_consumed = decode_note_on( tvb, pinfo, tree, cmd_count, offset, cmd_len, octet, *rsoffset, using_rs );
				break;
			case RTP_MIDI_STATUS_CHANNEL_POLYPHONIC_KEY_PRESSURE:
				ext_consumed = decode_poly_pressure(tvb, pinfo, tree, cmd_count, offset, cmd_len, octet, *rsoffset, using_rs );
				break;
			case RTP_MIDI_STATUS_CHANNEL_CONTROL_CHANGE:
				ext_consumed = decode_control_change(tvb, pinfo, tree, cmd_count, offset, cmd_len, octet, *rsoffset, using_rs );
				break;
			case RTP_MIDI_STATUS_CHANNEL_PROGRAM_CHANGE:
				ext_consumed = decode_program_change(tvb, pinfo, tree, cmd_count, offset, cmd_len, octet, *rsoffset, using_rs );
				break;
			case RTP_MIDI_STATUS_CHANNEL_CHANNEL_PRESSURE:
				ext_consumed = decode_channel_pressure(tvb, pinfo, tree, cmd_count, offset, cmd_len, octet, *rsoffset, using_rs );
				break;
			case RTP_MIDI_STATUS_CHANNEL_PITCH_BEND_CHANGE:
				ext_consumed = decode_pitch_bend_change(tvb, pinfo, tree, cmd_count, offset, cmd_len, octet, *rsoffset, using_rs );
				break;
		}
		/* external decoder informed us of error -> pass this through */
		if ( ext_consumed < 0 ) {
			return ext_consumed;
		}
		return consumed + ext_consumed;
	}

	/* Here we catch the remaining system-common commands */
	switch ( octet ) {
		case RTP_MIDI_STATUS_COMMON_SYSEX_START:
			ext_consumed =  decode_sysex_start( tvb, pinfo, tree, cmd_count, offset, cmd_len );
			break;
		case RTP_MIDI_STATUS_COMMON_MTC_QUARTER_FRAME:
			ext_consumed =  decode_mtc_quarter_frame( tvb, pinfo, tree, cmd_count, offset, cmd_len );
			break;
		case RTP_MIDI_STATUS_COMMON_SONG_POSITION_POINTER:
			ext_consumed =  decode_song_position_pointer( tvb, pinfo, tree, cmd_count, offset, cmd_len );
			break;
		case RTP_MIDI_STATUS_COMMON_SONG_SELECT:
			ext_consumed =  decode_song_select( tvb, pinfo, tree, cmd_count, offset, cmd_len );
			break;
		case RTP_MIDI_STATUS_COMMON_UNDEFINED_F4:
			ext_consumed =  decode_undefined_f4( tvb, pinfo, tree, cmd_count, offset, cmd_len );
			break;
		case RTP_MIDI_STATUS_COMMON_UNDEFINED_F5:
			ext_consumed =  decode_undefined_f5( tvb, pinfo, tree, cmd_count, offset, cmd_len );
			break;
		case RTP_MIDI_STATUS_COMMON_TUNE_REQUEST:
			ext_consumed =  decode_tune_request( tvb, pinfo, tree, cmd_count, offset, cmd_len );
			break;
		case RTP_MIDI_STATUS_COMMON_SYSEX_END:
			ext_consumed =  decode_sysex_end( tvb, pinfo, tree, cmd_count, offset, cmd_len );
			break;
	}
	/* external decoder informed us of error -> pass this through */
	if ( ext_consumed < 0 ) {
		return ext_consumed;
	}
	return consumed + ext_consumed;
}



/*
 * Here the chapter c of the channel-journal is decoded.
 */
static int
decode_cj_chapter_c( tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned int offset ) {
	proto_item			*ti				= NULL;
	proto_tree			*rtp_midi_cj_chapter_tree	= NULL;
	proto_tree			*rtp_midi_loglist_tree		= NULL;
	proto_tree			*rtp_midi_loglist_item_tree	= NULL;
	int				 consumed			= 0;
	guint8				 octet;
	int				 count;
	int				 i;

	/* Can we at least read the length of the control changes chapter? */
	if ( ( offset + 1 ) > tvb->length ) {
		return -1;
	}

	octet = tvb_get_guint8( tvb, offset );
	count = octet & 0x7f;

	/* count encoded is n+1 */
	count++;

	ti = proto_tree_add_text( tree, tvb, offset, 1 + 2 * count, RTP_MIDI_TREE_NAME_CJ_CHAPTER_C );
	rtp_midi_cj_chapter_tree = proto_item_add_subtree( ti, ett_rtp_midi_cj_chapter_c );

	proto_tree_add_item ( rtp_midi_cj_chapter_tree, hf_rtp_midi_cj_chapter_c_sflag, tvb, offset, 1, FALSE );
	proto_tree_add_item ( rtp_midi_cj_chapter_tree, hf_rtp_midi_cj_chapter_c_length, tvb, offset, 1, FALSE );

	offset++;
	consumed++;

	ti = proto_tree_add_text( rtp_midi_cj_chapter_tree, tvb, offset, count * 2, RTP_MIDI_TREE_NAME_CJ_CHAPTER_C_LOGLIST );
	rtp_midi_loglist_tree = proto_item_add_subtree( ti, ett_rtp_midi_cj_chapter_c_loglist );

	for ( i = 0; i < count; i++ ) {
		/* Can we still read this number/value-pair? */
		if ( ( offset + 2 ) > tvb->length ) {
			return -1;
		}

		ti = proto_tree_add_text(rtp_midi_loglist_tree, tvb, offset, 2, RTP_MIDI_TREE_NAME_CJ_CHAPTER_C_LOGITEM );
		ti = proto_item_add_subtree( ti, ett_rtp_midi_cj_chapter_c_logitem );

		proto_tree_add_item ( ti, hf_rtp_midi_cj_chapter_c_sflag, tvb, offset, 1, FALSE );
		proto_tree_add_item ( ti, hf_rtp_midi_cj_chapter_c_number, tvb, offset, 1, FALSE );

		offset++;
		consumed++;

		octet = tvb_get_guint8( tvb, offset );
		proto_tree_add_item ( ti, hf_rtp_midi_cj_chapter_c_aflag, tvb, offset, 1, FALSE );

		/* do we have a value or alternative coding? */
		if ( octet & 0x80 ) {
			proto_tree_add_item ( ti, hf_rtp_midi_cj_chapter_c_tflag, tvb, offset, 1, FALSE );
			proto_tree_add_item ( ti, hf_rtp_midi_cj_chapter_c_alt, tvb, offset, 1, FALSE );
		} else {
			proto_tree_add_item ( ti, hf_rtp_midi_cj_chapter_c_value, tvb, offset, 1, FALSE );
		}

		offset++;
		consumed++;
	}
	return consumed;
}



/*
 * Here the chapter m of the channel-journal is decoded, possibly the most complex part of the RTP-MIDI stuff ;-)
 */
static int
decode_cj_chapter_m( tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned int offset ) {
	proto_item			*ti				= NULL;
	proto_tree			*rtp_midi_cj_chapter_tree	= NULL;
	proto_tree			*rtp_midi_loglist_tree		= NULL;
	proto_tree			*rtp_midi_loglist_item_tree	= NULL;
	guint16				header;
	guint8				logitemheader;
	int				length;
	int				logitemhdrlen;
	int				logitemlen;
	gboolean			no_pnum_msb;
	int				consumed = 0;

	/* Can we read this channel-journals header? */
	if ( ( offset + 2 ) > tvb->length ) {
		return -1;
	}

	/* first we need to get the flags & length of this chapter */
	header = tvb_get_ntohs( tvb, offset );
	length = header & RTP_MIDI_CJ_CHAPTER_M_MASK_LENGTH;
	/* take of length of header */
	length -= 2;

	ti = proto_tree_add_text( tree, tvb, offset, length, RTP_MIDI_TREE_NAME_CJ_CHAPTER_M );
	rtp_midi_cj_chapter_tree = proto_item_add_subtree( ti, ett_rtp_midi_cj_chapter_m );

	proto_tree_add_item ( rtp_midi_cj_chapter_tree, hf_rtp_midi_cj_chapter_m_sflag, tvb, offset, 2, FALSE );
	proto_tree_add_item ( rtp_midi_cj_chapter_tree, hf_rtp_midi_cj_chapter_m_pflag, tvb, offset, 2, FALSE );
	proto_tree_add_item ( rtp_midi_cj_chapter_tree, hf_rtp_midi_cj_chapter_m_eflag, tvb, offset, 2, FALSE );
	proto_tree_add_item ( rtp_midi_cj_chapter_tree, hf_rtp_midi_cj_chapter_m_uflag, tvb, offset, 2, FALSE );
	proto_tree_add_item ( rtp_midi_cj_chapter_tree, hf_rtp_midi_cj_chapter_m_wflag, tvb, offset, 2, FALSE );
	proto_tree_add_item ( rtp_midi_cj_chapter_tree, hf_rtp_midi_cj_chapter_m_zflag, tvb, offset, 2, FALSE );
	proto_tree_add_item ( rtp_midi_cj_chapter_tree, hf_rtp_midi_cj_chapter_m_length, tvb, offset, 2, FALSE );

	/* done with header */
	consumed += 2;
	offset += 2;

	/* do we have the pending field? */
	if ( header & 0x4000 ) {
		/* Can we read this pending field? */
		if ( ( offset + 1 ) > tvb->length ) {
			return -1;
		}

		proto_tree_add_item ( rtp_midi_cj_chapter_tree, hf_rtp_midi_cj_chapter_m_qflag, tvb, offset, 1, FALSE );
		proto_tree_add_item ( rtp_midi_cj_chapter_tree, hf_rtp_midi_cj_chapter_m_pending, tvb, offset, 1, FALSE );

		consumed++;
		offset++;
	}

	/*
	 * lets find out if we need to decode the pnum_msb:
	 * if Z = 1 and either U = 1 or W = 1 we don't
	 */
	no_pnum_msb = ( header & 0x0400 ) && ( ( header & 0x0800 ) || ( header & 0x1000 ) );
	logitemhdrlen = no_pnum_msb ? 2 : 3;

	ti = proto_tree_add_text( rtp_midi_cj_chapter_tree, tvb, offset, length, RTP_MIDI_TREE_NAME_CJ_CHAPTER_M_LOGLIST );
	rtp_midi_loglist_tree = proto_item_add_subtree( ti, ett_rtp_midi_cj_chapter_m_loglist );

	/* lets step through the loglist */
	while ( length > 0 ) {

		/* Can we read this channel-journals header? */
		if ( ( offset + logitemhdrlen ) > tvb->length ) {
			return -1;
		}

		if ( no_pnum_msb ) {
			logitemheader = tvb_get_guint8( tvb, offset + 1 );
		} else {
			logitemheader = tvb_get_guint8( tvb, offset + 1 );
		}

		logitemlen = logitemhdrlen;

		/* do we have a msb field? */
		if ( logitemheader & RTP_MIDI_CJ_CHAPTER_M_FLAG_J ) {
			logitemlen++;
		}

		/* do we have a lsb field? */
		if ( logitemheader & RTP_MIDI_CJ_CHAPTER_M_FLAG_K ) {
			logitemlen++;
		}

		/* do we have an a-button field? */
		if ( logitemheader & RTP_MIDI_CJ_CHAPTER_M_FLAG_L ) {
			logitemlen +=2;
		}

		/* do we have a c-button field? */
		if ( logitemheader & RTP_MIDI_CJ_CHAPTER_M_FLAG_M ) {
			logitemlen +=2;
		}

		/* do we have a count field? */
		if ( logitemheader & RTP_MIDI_CJ_CHAPTER_M_FLAG_N ) {
			logitemlen++;
		}

		/* now that we have the length of the logitem, we can create the tree... */
		ti = proto_tree_add_text( rtp_midi_loglist_tree, tvb, offset, logitemlen, RTP_MIDI_TREE_NAME_CJ_CHAPTER_M_LOGITEM );
		rtp_midi_loglist_item_tree = proto_item_add_subtree( ti, ett_rtp_midi_cj_chapter_m_logitem );

		proto_tree_add_item ( rtp_midi_loglist_item_tree, hf_rtp_midi_cj_chapter_m_log_sflag, tvb, offset, 1, FALSE );
		proto_tree_add_item ( rtp_midi_loglist_item_tree, hf_rtp_midi_cj_chapter_m_log_pnum_lsb, tvb, offset, 1, FALSE );
		consumed++;
		offset++;
		length--;

		/* if we have the msb, we need to decode it */
		if ( !no_pnum_msb ) {
			proto_tree_add_item ( rtp_midi_loglist_item_tree, hf_rtp_midi_cj_chapter_m_log_qflag, tvb, offset, 1, FALSE );
			proto_tree_add_item ( rtp_midi_loglist_item_tree, hf_rtp_midi_cj_chapter_m_log_pnum_msb, tvb, offset, 1, FALSE );
			consumed++;
			offset++;
			length--;
		}

		proto_tree_add_item ( rtp_midi_loglist_item_tree, hf_rtp_midi_cj_chapter_m_log_jflag, tvb, offset, 1, FALSE );
		proto_tree_add_item ( rtp_midi_loglist_item_tree, hf_rtp_midi_cj_chapter_m_log_kflag, tvb, offset, 1, FALSE );
		proto_tree_add_item ( rtp_midi_loglist_item_tree, hf_rtp_midi_cj_chapter_m_log_lflag, tvb, offset, 1, FALSE );
		proto_tree_add_item ( rtp_midi_loglist_item_tree, hf_rtp_midi_cj_chapter_m_log_mflag, tvb, offset, 1, FALSE );
		proto_tree_add_item ( rtp_midi_loglist_item_tree, hf_rtp_midi_cj_chapter_m_log_nflag, tvb, offset, 1, FALSE );
		proto_tree_add_item ( rtp_midi_loglist_item_tree, hf_rtp_midi_cj_chapter_m_log_tflag, tvb, offset, 1, FALSE );
		proto_tree_add_item ( rtp_midi_loglist_item_tree, hf_rtp_midi_cj_chapter_m_log_vflag, tvb, offset, 1, FALSE );
		proto_tree_add_item ( rtp_midi_loglist_item_tree, hf_rtp_midi_cj_chapter_m_log_rflag, tvb, offset, 1, FALSE );
		consumed++;
		offset++;
		length--;

		/* do we have a msb field? */
		if ( logitemheader & RTP_MIDI_CJ_CHAPTER_M_FLAG_J ) {

			/* Can we read this table? */
			if ( ( offset + 1 ) > tvb->length ) {
				return -1;
			}

			ti = proto_tree_add_text( rtp_midi_loglist_tree, tvb, offset, 1, RTP_MIDI_TREE_NAME_CJ_CHAPTER_M_LOG_MSB );
			ti = proto_item_add_subtree( ti, ett_rtp_midi_cj_chapter_m_log_msb );

			proto_tree_add_item ( ti, hf_rtp_midi_cj_chapter_m_log_msb_x, tvb, offset, 1, FALSE );
			proto_tree_add_item ( ti, hf_rtp_midi_cj_chapter_m_log_msb, tvb, offset, 1, FALSE );

			consumed++;
			offset++;
			length--;
		}

		/* do we have a lsb field? */
		if ( logitemheader & RTP_MIDI_CJ_CHAPTER_M_FLAG_K ) {
			/* Can we read this table? */
			if ( ( offset + 1 ) > tvb->length ) {
				return -1;
			}

			ti = proto_tree_add_text( rtp_midi_loglist_tree, tvb, offset, 1, RTP_MIDI_TREE_NAME_CJ_CHAPTER_M_LOG_LSB );
			ti = proto_item_add_subtree( ti, ett_rtp_midi_cj_chapter_m_log_lsb );

			proto_tree_add_item ( ti, hf_rtp_midi_cj_chapter_m_log_lsb_x, tvb, offset, 1, FALSE );
			proto_tree_add_item ( ti, hf_rtp_midi_cj_chapter_m_log_lsb, tvb, offset, 1, FALSE );

			consumed++;
			offset++;
			length--;
		}

		/* do we have an a-button field? */
		if ( logitemheader & RTP_MIDI_CJ_CHAPTER_M_FLAG_L ) {
			/* Can we read this table? */
			if ( ( offset + 2 ) > tvb->length ) {
				return -1;
			}

			ti = proto_tree_add_text( rtp_midi_loglist_tree, tvb, offset, 2, RTP_MIDI_TREE_NAME_CJ_CHAPTER_M_LOG_A_BUTTON );
			ti = proto_item_add_subtree( ti, ett_rtp_midi_cj_chapter_m_log_a_button );

			proto_tree_add_item ( ti, hf_rtp_midi_cj_chapter_m_log_a_button_g, tvb, offset, 2, FALSE );
			proto_tree_add_item ( ti, hf_rtp_midi_cj_chapter_m_log_a_button_x, tvb, offset, 2, FALSE );
			proto_tree_add_item ( ti, hf_rtp_midi_cj_chapter_m_log_a_button, tvb, offset, 2, FALSE );

			consumed += 2;
			offset += 2;
			length -= 2;
		}

		/* do we have a c-button field? */
		if ( logitemheader & RTP_MIDI_CJ_CHAPTER_M_FLAG_M ) {
			/* Can we read this table? */
			if ( ( offset + 2 ) > tvb->length ) {
				return -1;
			}
			ti = proto_tree_add_text( rtp_midi_loglist_tree, tvb, offset, 2, RTP_MIDI_TREE_NAME_CJ_CHAPTER_M_LOG_C_BUTTON );
			ti = proto_item_add_subtree( ti, ett_rtp_midi_cj_chapter_m_log_c_button );

			proto_tree_add_item ( ti, hf_rtp_midi_cj_chapter_m_log_c_button_g, tvb, offset, 2, FALSE );
			proto_tree_add_item ( ti, hf_rtp_midi_cj_chapter_m_log_c_button_r, tvb, offset, 2, FALSE );
			proto_tree_add_item ( ti, hf_rtp_midi_cj_chapter_m_log_c_button, tvb, offset, 2, FALSE );

			consumed += 2;
			offset += 2;
			length -= 2;
		}

		/* do we have a count field? */
		if ( logitemheader & RTP_MIDI_CJ_CHAPTER_M_FLAG_N ) {
			/* Can we read this table? */
			if ( ( offset + 1 ) > tvb->length ) {
				return -1;
			}
			ti = proto_tree_add_text( rtp_midi_loglist_tree, tvb, offset, 1, RTP_MIDI_TREE_NAME_CJ_CHAPTER_M_LOG_COUNT );
			ti = proto_item_add_subtree( ti, ett_rtp_midi_cj_chapter_m_log_count );

			proto_tree_add_item ( ti, hf_rtp_midi_cj_chapter_m_log_count_x, tvb, offset, 1, FALSE );
			proto_tree_add_item ( ti, hf_rtp_midi_cj_chapter_m_log_count, tvb, offset, 1, FALSE );

			consumed++;
			offset++;
			length--;
		}

	}


	return consumed;
}


/*
 * Here the chapter n of the channel-journal is decoded.
 */
static int
decode_cj_chapter_n( tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned int offset ) {
	proto_item			*ti				= NULL;
	proto_tree			*rtp_midi_cj_chapter_tree	= NULL;
	proto_tree			*rtp_midi_loglist_tree		= NULL;
	proto_tree			*rtp_midi_loglist_item_tree	= NULL;
	const gchar			*note_str			= NULL;
	int				 consumed			= 0;
	guint16				 header;
	guint8				 note;
	guint8				 velocity;
	int				 log_count;
	int				 octet_count;
	int				 low;
	int				 high;
	int				 i;

	/* Can we read this channel-journals header? */
	if ( ( offset + 2 ) > tvb->length ) {
		return -1;
	}

	/* first we need to get the flags & length of this chapter */
	header = tvb_get_ntohs( tvb, offset );
	log_count = ( header & RTP_MIDI_CJ_CHAPTER_N_MASK_LENGTH ) >> 8;
	low = ( header & RTP_MIDI_CJ_CHAPTER_N_MASK_LOW ) >> 4;
	high = header & RTP_MIDI_CJ_CHAPTER_N_MASK_HIGH;

	/* how many offbits octets do we have? */
	if ( low <= high ) {
		octet_count = high - low + 1;
	} else if ( ( low == 15 ) && ( high == 0 ) ) {
		octet_count = 0;
	} else if ( ( low == 15 ) && ( high == 1 ) ) {
		octet_count = 0;
	} else {
		return -1;
	}

	/* special case -> no offbit octets, but 128 note-logs */
	if ( ( log_count == 127 ) && ( low == 15) && ( high == 0 ) ) {
		log_count++;
	}

	ti = proto_tree_add_text( tree, tvb, offset, 2 + ( log_count * 2 ) + octet_count, RTP_MIDI_TREE_NAME_CJ_CHAPTER_N );
	rtp_midi_cj_chapter_tree = proto_item_add_subtree( ti, ett_rtp_midi_cj_chapter_n );

	proto_tree_add_item ( rtp_midi_cj_chapter_tree, hf_rtp_midi_cj_chapter_n_bflag, tvb, offset, 2, FALSE );
	proto_tree_add_item ( rtp_midi_cj_chapter_tree, hf_rtp_midi_cj_chapter_n_len, tvb, offset, 2, FALSE );
	proto_tree_add_item ( rtp_midi_cj_chapter_tree, hf_rtp_midi_cj_chapter_n_low, tvb, offset, 2, FALSE );
	proto_tree_add_item ( rtp_midi_cj_chapter_tree, hf_rtp_midi_cj_chapter_n_high, tvb, offset, 2, FALSE );

	offset += 2;
	consumed += 2;

	if ( log_count > 0 ) {

		/* Can we read this loglist? */
		if ( ( offset + ( log_count * 2 ) ) > tvb->length ) {
			return -1;
		}

		ti = proto_tree_add_text( rtp_midi_cj_chapter_tree, tvb, offset, log_count * 2, RTP_MIDI_TREE_NAME_CJ_CHAPTER_N_LOGLIST );
		rtp_midi_loglist_tree = proto_item_add_subtree( ti, ett_rtp_midi_cj_chapter_n_loglist );

		for ( i = 0; i < log_count; i++ ) {

			note = tvb_get_guint8( tvb, offset ) & 0x7f;
			velocity = tvb_get_guint8( tvb, offset + 1 ) & 0x7f;;

			note_str = val_to_str( note, rtp_midi_note_values, "" );

			ti = proto_tree_add_text(rtp_midi_loglist_tree, tvb, offset, 2, "%s (n=%s, v=%d)", RTP_MIDI_TREE_NAME_CJ_CHAPTER_N_LOGITEM, note_str, velocity );
			ti = proto_item_add_subtree( ti, ett_rtp_midi_cj_chapter_n_logitem );

			proto_tree_add_item ( ti, hf_rtp_midi_cj_chapter_n_log_sflag, tvb, offset, 1, FALSE );
			proto_tree_add_item ( ti, hf_rtp_midi_cj_chapter_n_log_notenum, tvb, offset, 1, FALSE );

			offset++;
			consumed++;

			proto_tree_add_item ( ti, hf_rtp_midi_cj_chapter_n_log_yflag, tvb, offset, 1, FALSE );
			proto_tree_add_item ( ti, hf_rtp_midi_cj_chapter_n_log_velocity, tvb, offset, 1, FALSE );

			offset++;
			consumed++;
		}

	}

	if ( octet_count > 0 ) {

		/* Can we read this loglist? */
		if ( ( offset + octet_count ) > tvb->length ) {
			return -1;
		}
		ti = proto_tree_add_text( rtp_midi_cj_chapter_tree, tvb, offset, log_count, RTP_MIDI_TREE_NAME_CJ_CHAPTER_N_OCTETS );
		rtp_midi_loglist_tree = proto_item_add_subtree( ti, ett_rtp_midi_cj_chapter_n_octets );

		for ( i = 0; i < octet_count; i++ ) {

			proto_tree_add_item ( ti, hf_rtp_midi_cj_chapter_n_log_octet, tvb, offset, 1, FALSE );

			offset++;
			consumed++;
		}
	}

	return consumed;
}


/*
 * Here the chapter e of the channel-journal is decoded.
 */
static int
decode_cj_chapter_e( tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned int offset ) {
	proto_item			*ti				= NULL;
	proto_tree			*rtp_midi_cj_chapter_tree	= NULL;
	proto_tree			*rtp_midi_loglist_tree		= NULL;
	proto_tree			*rtp_midi_loglist_item_tree	= NULL;
	const gchar			*note_str			= NULL;
	int				 consumed			= 0;
	guint8				 header;
	guint8				 note;
	guint8				 count_vel;
	guint8				 octet;
	int				 log_count;
	int				 i;

	/* Can we read this channel-journals header? */
	if ( ( offset + 1 ) > tvb->length ) {
		return -1;
	}

	/* first we need to get the flags & length of this chapter */
	header = tvb_get_guint8( tvb, offset );
	log_count = header & RTP_MIDI_CJ_CHAPTER_E_MASK_LENGTH;

	/* count is encoded n+1 */
	log_count++;

	ti = proto_tree_add_text( tree, tvb, offset, 1 + ( log_count * 2 ), RTP_MIDI_TREE_NAME_CJ_CHAPTER_E );
	rtp_midi_cj_chapter_tree = proto_item_add_subtree( ti, ett_rtp_midi_cj_chapter_e );

	proto_tree_add_item ( rtp_midi_cj_chapter_tree, hf_rtp_midi_cj_chapter_e_sflag, tvb, offset, 1, FALSE );
	proto_tree_add_item ( rtp_midi_cj_chapter_tree, hf_rtp_midi_cj_chapter_e_len, tvb, offset, 2, FALSE );

	offset++;
	consumed++;

	/* Can we read this loglist? */
	if ( ( offset + ( log_count * 2 ) ) > tvb->length ) {
		return -1;
	}

	ti = proto_tree_add_text( rtp_midi_cj_chapter_tree, tvb, offset, log_count * 2, RTP_MIDI_TREE_NAME_CJ_CHAPTER_E_LOGLIST );
	rtp_midi_loglist_tree = proto_item_add_subtree( ti, ett_rtp_midi_cj_chapter_e_loglist );

	for ( i = 0; i < log_count; i++ ) {

		note = tvb_get_guint8( tvb, offset ) & 0x7f;
		octet = tvb_get_guint8( tvb, offset + 1 );
		count_vel = octet & 0x7f;

		note_str = val_to_str( note, rtp_midi_note_values, "" );

		if ( octet & 0x80 ) {
			ti = proto_tree_add_text(rtp_midi_loglist_tree, tvb, offset, 2, "%s (n=%s, v=%d)", RTP_MIDI_TREE_NAME_CJ_CHAPTER_E_LOGITEM1, note_str, count_vel );
		} else {
			ti = proto_tree_add_text(rtp_midi_loglist_tree, tvb, offset, 2, "%s (n=%s, c=%d)", RTP_MIDI_TREE_NAME_CJ_CHAPTER_E_LOGITEM2, note_str, count_vel );
		}
		ti = proto_item_add_subtree( ti, ett_rtp_midi_cj_chapter_e_logitem );

		proto_tree_add_item ( ti, hf_rtp_midi_cj_chapter_e_log_sflag, tvb, offset, 1, FALSE );
		proto_tree_add_item ( ti, hf_rtp_midi_cj_chapter_e_log_notenum, tvb, offset, 1, FALSE );

		offset++;
		consumed++;

		proto_tree_add_item ( ti, hf_rtp_midi_cj_chapter_e_log_vflag, tvb, offset, 1, FALSE );
		if ( octet & 0x80 ) {
			proto_tree_add_item ( ti, hf_rtp_midi_cj_chapter_e_log_velocity, tvb, offset, 1, FALSE );
		} else {
			proto_tree_add_item ( ti, hf_rtp_midi_cj_chapter_e_log_count, tvb, offset, 1, FALSE );
		}

		offset++;
		consumed++;
	}

	return consumed;
}








/*
 * Here the chapter a of the channel-journal is decoded.
 */
static int
decode_cj_chapter_a( tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned int offset ) {
	proto_item			*ti				= NULL;
	proto_tree			*rtp_midi_cj_chapter_tree	= NULL;
	proto_tree			*rtp_midi_loglist_tree		= NULL;
	proto_tree			*rtp_midi_loglist_item_tree	= NULL;
	const gchar			*note_str			= NULL;
	int				 consumed			= 0;
	guint8				 header;
	guint8				 note;
	guint8				 pressure;
	int				 log_count;
	int				 i;

	/* Can we read this channel-journals header? */
	if ( ( offset + 1 ) > tvb->length ) {
		return -1;
	}

	/* first we need to get the flags & length of this chapter */
	header = tvb_get_guint8( tvb, offset );
	log_count = header & RTP_MIDI_CJ_CHAPTER_A_MASK_LENGTH;

	/* count is encoded n+1 */
	log_count++;

	ti = proto_tree_add_text( tree, tvb, offset, 1 + ( log_count * 2 ), RTP_MIDI_TREE_NAME_CJ_CHAPTER_A );
	rtp_midi_cj_chapter_tree = proto_item_add_subtree( ti, ett_rtp_midi_cj_chapter_a );

	proto_tree_add_item ( rtp_midi_cj_chapter_tree, hf_rtp_midi_cj_chapter_a_sflag, tvb, offset, 1, FALSE );
	proto_tree_add_item ( rtp_midi_cj_chapter_tree, hf_rtp_midi_cj_chapter_a_len, tvb, offset, 2, FALSE );

	offset++;
	consumed++;

	/* Can we read this loglist? */
	if ( ( offset + ( log_count * 2 ) ) > tvb->length ) {
		return -1;
	}

	ti = proto_tree_add_text( rtp_midi_cj_chapter_tree, tvb, offset, log_count * 2, RTP_MIDI_TREE_NAME_CJ_CHAPTER_A_LOGLIST );
	rtp_midi_loglist_tree = proto_item_add_subtree( ti, ett_rtp_midi_cj_chapter_a_loglist );

	for ( i = 0; i < log_count; i++ ) {

		note = tvb_get_guint8( tvb, offset ) & 0x7f;
		pressure = tvb_get_guint8( tvb, offset + 1 ) & 0x7f;

		note_str = val_to_str( note, rtp_midi_note_values, "" );

		ti = proto_tree_add_text(rtp_midi_loglist_tree, tvb, offset, 2, "%s (n=%s, p=%d)", RTP_MIDI_TREE_NAME_CJ_CHAPTER_A_LOGITEM, note_str, pressure );
		ti = proto_item_add_subtree( ti, ett_rtp_midi_cj_chapter_a_logitem );

		proto_tree_add_item ( ti, hf_rtp_midi_cj_chapter_a_log_sflag, tvb, offset, 1, FALSE );
		proto_tree_add_item ( ti, hf_rtp_midi_cj_chapter_a_log_notenum, tvb, offset, 1, FALSE );

		offset++;
		consumed++;

		proto_tree_add_item ( ti, hf_rtp_midi_cj_chapter_a_log_xflag, tvb, offset, 1, FALSE );
		proto_tree_add_item ( ti, hf_rtp_midi_cj_chapter_a_log_pressure, tvb, offset, 1, FALSE );

		offset++;
		consumed++;
	}

	return consumed;
}





/*
 * Here a channel-journal is decoded.
 */
static int
decode_channel_journal( tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned int offset ) {
	proto_item			*ti				= NULL;
	proto_tree			*rtp_midi_chanjournal_tree	= NULL;
	proto_tree			*rtp_midi_cj_chapters_tree	= NULL;
	proto_tree			*rtp_midi_cj_chapter_tree	= NULL;
	guint32				chanflags;
	guint16				chanjourlen;
	int				consumed = 0;
	int				ext_consumed = 0;

	/* Can we read this channel-journals header? */
	if ( ( offset + 3 ) > tvb->length ) {
		return -1;
	}

	/* first we need to get the flags & length of this channel-journal */
	chanflags = tvb_get_ntoh24( tvb, offset );
	chanjourlen = ( chanflags & RTP_MIDI_CJ_MASK_LENGTH ) >> 8;

	ti = proto_tree_add_text( tree, tvb, offset, chanjourlen, val_to_str( ( chanflags & RTP_MIDI_CJ_MASK_CHANNEL ) >> RTP_MIDI_CJ_CHANNEL_SHIFT, rtp_midi_channels, "" ) );
	rtp_midi_chanjournal_tree = proto_item_add_subtree( ti, ett_rtp_midi_channeljournal );

	proto_tree_add_item ( rtp_midi_chanjournal_tree, hf_rtp_midi_chanjour_sflag, tvb, offset, 3, FALSE );
	proto_tree_add_item ( rtp_midi_chanjournal_tree, hf_rtp_midi_chanjour_chan, tvb, offset, 3, FALSE );
	proto_tree_add_item ( rtp_midi_chanjournal_tree, hf_rtp_midi_chanjour_hflag, tvb, offset, 3, FALSE );
	proto_tree_add_item ( rtp_midi_chanjournal_tree, hf_rtp_midi_chanjour_len, tvb, offset, 3, FALSE );
	proto_tree_add_item ( rtp_midi_chanjournal_tree, hf_rtp_midi_chanjour_toc_p, tvb, offset, 3, FALSE );
	proto_tree_add_item ( rtp_midi_chanjournal_tree, hf_rtp_midi_chanjour_toc_c, tvb, offset, 3, FALSE );
	proto_tree_add_item ( rtp_midi_chanjournal_tree, hf_rtp_midi_chanjour_toc_m, tvb, offset, 3, FALSE );
	proto_tree_add_item ( rtp_midi_chanjournal_tree, hf_rtp_midi_chanjour_toc_w, tvb, offset, 3, FALSE );
	proto_tree_add_item ( rtp_midi_chanjournal_tree, hf_rtp_midi_chanjour_toc_n, tvb, offset, 3, FALSE );
	proto_tree_add_item ( rtp_midi_chanjournal_tree, hf_rtp_midi_chanjour_toc_e, tvb, offset, 3, FALSE );
	proto_tree_add_item ( rtp_midi_chanjournal_tree, hf_rtp_midi_chanjour_toc_t, tvb, offset, 3, FALSE );
	proto_tree_add_item ( rtp_midi_chanjournal_tree, hf_rtp_midi_chanjour_toc_a, tvb, offset, 3, FALSE );

	ti = proto_tree_add_text( rtp_midi_chanjournal_tree, tvb, offset + 3, chanjourlen - 3, RTP_MIDI_TREE_NAME_CHANNEL_CHAPTERS );
	rtp_midi_cj_chapters_tree = proto_item_add_subtree( ti, ett_rtp_midi_channelchapters );

	/* take care of length of header */
	offset += 3;
	consumed += 3;

	/* Do we have a program change chapter? */
	if ( chanflags & RTP_MIDI_CJ_FLAG_P ) {

		/* Can we read the program change chapter? */
		if ( ( offset + 3 ) > tvb->length ) {
			return -1;
		}

		ti = proto_tree_add_text( rtp_midi_cj_chapters_tree, tvb, offset, 3, RTP_MIDI_TREE_NAME_CJ_CHAPTER_P );
		rtp_midi_cj_chapter_tree = proto_item_add_subtree( ti, ett_rtp_midi_cj_chapter_p );

		proto_tree_add_item ( rtp_midi_cj_chapter_tree, hf_rtp_midi_cj_chapter_p_sflag, tvb, offset, 1, FALSE );
		proto_tree_add_item ( rtp_midi_cj_chapter_tree, hf_rtp_midi_cj_chapter_p_program, tvb, offset, 1, FALSE );
		proto_tree_add_item ( rtp_midi_cj_chapter_tree, hf_rtp_midi_cj_chapter_p_bflag, tvb, offset + 1, 1, FALSE );
		proto_tree_add_item ( rtp_midi_cj_chapter_tree, hf_rtp_midi_cj_chapter_p_bank_msb, tvb, offset + 1, 1, FALSE );
		proto_tree_add_item ( rtp_midi_cj_chapter_tree, hf_rtp_midi_cj_chapter_p_xflag, tvb, offset + 2, 1, FALSE );
		proto_tree_add_item ( rtp_midi_cj_chapter_tree, hf_rtp_midi_cj_chapter_p_bank_lsb, tvb, offset + 2, 1, FALSE );

		offset += 3;
		consumed += 3;
	}

	/* Do we have a control chapter? */
	if ( chanflags & RTP_MIDI_CJ_FLAG_C ) {
		ext_consumed = decode_cj_chapter_c( tvb, pinfo, rtp_midi_cj_chapters_tree, offset );
		if ( ext_consumed < 0 ) {
			return ext_consumed;
		}
		consumed += ext_consumed;
		offset += ext_consumed;

	}

	/* Do we have a parameter changes? */
	if ( chanflags & RTP_MIDI_CJ_FLAG_M ) {
		ext_consumed = decode_cj_chapter_m( tvb, pinfo, rtp_midi_cj_chapters_tree, offset );
		if ( ext_consumed < 0 ) {
			return ext_consumed;
		}
		consumed += ext_consumed;
		offset += ext_consumed;
	}

	/* Do we have a pitch-wheel chapter? */
	if ( chanflags & RTP_MIDI_CJ_FLAG_W ) {

		/* Can we get the data for the Pitch Wheel chapter? */
		if ( ( offset + 2 ) > tvb->length ) {
			return -1;
		}

		ti = proto_tree_add_text( rtp_midi_cj_chapters_tree, tvb, offset, 2, RTP_MIDI_TREE_NAME_CJ_CHAPTER_W );
		rtp_midi_cj_chapter_tree = proto_item_add_subtree( ti, ett_rtp_midi_cj_chapter_w );

		proto_tree_add_item ( rtp_midi_cj_chapter_tree, hf_rtp_midi_cj_chapter_w_sflag, tvb, offset, 1, FALSE );
		proto_tree_add_item ( rtp_midi_cj_chapter_tree, hf_rtp_midi_cj_chapter_w_first, tvb, offset, 1, FALSE );

		offset++;
		consumed++;

		proto_tree_add_item ( rtp_midi_cj_chapter_tree, hf_rtp_midi_cj_chapter_w_rflag, tvb, offset, 1, FALSE );
		proto_tree_add_item ( rtp_midi_cj_chapter_tree, hf_rtp_midi_cj_chapter_w_second, tvb, offset, 1, FALSE );

		offset++;
		consumed++;
	}

	/* Do we have a note on/off chapter? */
	if ( chanflags & RTP_MIDI_CJ_FLAG_N ) {
		ext_consumed = decode_cj_chapter_n( tvb, pinfo, rtp_midi_cj_chapters_tree, offset );
		if ( ext_consumed < 0 ) {
			return ext_consumed;
		}
		consumed += ext_consumed;
		offset += ext_consumed;
	}

	/* Do we have a note command extras chapter? */
	if ( chanflags & RTP_MIDI_CJ_FLAG_E ) {
		ext_consumed = decode_cj_chapter_e( tvb, pinfo, rtp_midi_cj_chapters_tree, offset );
		if ( ext_consumed < 0 ) {
			return ext_consumed;
		}
		consumed += ext_consumed;
		offset += ext_consumed;
	}

	/* Do we have channel aftertouch chapter? */
	if ( chanflags & RTP_MIDI_CJ_FLAG_T ) {
		/* Can we get the data for the Pitch Wheel chapter? */
		if ( ( offset + 1 ) > tvb->length ) {
			return -1;
		}

		ti = proto_tree_add_text( rtp_midi_cj_chapters_tree, tvb, offset, 1, RTP_MIDI_TREE_NAME_CJ_CHAPTER_T );
		rtp_midi_cj_chapter_tree = proto_item_add_subtree( ti, ett_rtp_midi_cj_chapter_t );

		proto_tree_add_item ( rtp_midi_cj_chapter_tree, hf_rtp_midi_cj_chapter_t_sflag, tvb, offset, 1, FALSE );
		proto_tree_add_item ( rtp_midi_cj_chapter_tree, hf_rtp_midi_cj_chapter_t_pressure, tvb, offset, 1, FALSE );

		offset++;
		consumed++;
	}

	/* Do we have a poly aftertouch chapter? */
	if ( chanflags & RTP_MIDI_CJ_FLAG_A ) {
		ext_consumed = decode_cj_chapter_a( tvb, pinfo, rtp_midi_cj_chapters_tree, offset );
		if ( ext_consumed < 0 ) {
			return ext_consumed;
		}
		consumed += ext_consumed;
		offset += ext_consumed;
	}

	/* Make sanity check for consumed data vs. stated length of this channels journal */
	if ( consumed != chanjourlen ) {
		return -1;
	}
	return consumed;
}


/*
 * Here the chapter D F4-field of the system-journal is decoded.
 */
static int
decode_sj_chapter_d_f4( tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned int offset ) {
	return -1;
}


/*
 * Here the chapter D F5-field of the system-journal is decoded.
 */
static int
decode_sj_chapter_d_f5( tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned int offset ) {
	return -1;
}

/*
 * Here the chapter D F9-field of the system-journal is decoded.
 */
static int
decode_sj_chapter_d_f9( tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned int offset ) {
	return -1;
}


/*
 * Here the chapter D FD-field of the system-journal is decoded.
 */
static int
decode_sj_chapter_d_fd( tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned int offset ) {
	return -1;
}



/*
 * Here the chapter D of the channel-journal is decoded.
 */
static int
decode_sj_chapter_d( tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned int offset ) {
	proto_item			*ti				= NULL;
	proto_item			*tix				= NULL;
	proto_tree			*rtp_midi_sj_chapter_tree	= NULL;
	proto_tree			*rtp_midi_field_tree		= NULL;
	proto_tree			*rtp_midi_loglist_item_tree	= NULL;
	guint8				header;
	int				consumed = 0;
	int				ext_consumed;

	/* Can we read this channel-journals header? */
	if ( ( offset + 1 ) > tvb->length ) {
		return -1;
	}

	/* first we need to get the flags of this chapter */
	header = tvb_get_guint8( tvb, offset );

	/*
	 * we don't know the complete length yet - we now that we have at least the header -> 1.
	 * At the end we set the real length...
	 */
	tix = proto_tree_add_text( tree, tvb, offset, 1, RTP_MIDI_TREE_NAME_SJ_CHAPTER_D );
	rtp_midi_sj_chapter_tree = proto_item_add_subtree( tix, ett_rtp_midi_sj_chapter_d );

	proto_tree_add_item ( rtp_midi_sj_chapter_tree, hf_rtp_midi_sj_chapter_d_sflag, tvb, offset, 1, FALSE );
	proto_tree_add_item ( rtp_midi_sj_chapter_tree, hf_rtp_midi_sj_chapter_d_bflag, tvb, offset, 1, FALSE );
	proto_tree_add_item ( rtp_midi_sj_chapter_tree, hf_rtp_midi_sj_chapter_d_gflag, tvb, offset, 1, FALSE );
	proto_tree_add_item ( rtp_midi_sj_chapter_tree, hf_rtp_midi_sj_chapter_d_hflag, tvb, offset, 1, FALSE );
	proto_tree_add_item ( rtp_midi_sj_chapter_tree, hf_rtp_midi_sj_chapter_d_jflag, tvb, offset, 1, FALSE );
	proto_tree_add_item ( rtp_midi_sj_chapter_tree, hf_rtp_midi_sj_chapter_d_kflag, tvb, offset, 1, FALSE );
	proto_tree_add_item ( rtp_midi_sj_chapter_tree, hf_rtp_midi_sj_chapter_d_yflag, tvb, offset, 1, FALSE );
	proto_tree_add_item ( rtp_midi_sj_chapter_tree, hf_rtp_midi_sj_chapter_d_zflag, tvb, offset, 1, FALSE );

	/* done with header */
	consumed++;
	offset++;

	/* do we have Reset field? */
	if ( header & RTP_MIDI_SJ_CHAPTER_D_FLAG_B ) {
		/* Can we get the data for the Reset chapter? */
		if ( ( offset + 1 ) > tvb->length ) {
			return -1;
		}

		ti = proto_tree_add_text( rtp_midi_sj_chapter_tree, tvb, offset, 1, RTP_MIDI_TREE_NAME_SJ_CHAPTER_D_FIELD_B );
		rtp_midi_field_tree = proto_item_add_subtree( ti, ett_rtp_midi_sj_chapter_d_field_b );

		proto_tree_add_item ( rtp_midi_field_tree, hf_rtp_midi_sj_chapter_d_reset_sflag, tvb, offset, 1, FALSE );
		proto_tree_add_item ( rtp_midi_field_tree, hf_rtp_midi_sj_chapter_d_reset_count, tvb, offset, 1, FALSE );

		offset++;
		consumed++;
	}

	/* do we have Tune request field? */
	if ( header & RTP_MIDI_SJ_CHAPTER_D_FLAG_G ) {
		/* Can we get the data for the Reset chapter? */
		if ( ( offset + 1 ) > tvb->length ) {
			return -1;
		}

		ti = proto_tree_add_text( rtp_midi_sj_chapter_tree, tvb, offset, 1, RTP_MIDI_TREE_NAME_SJ_CHAPTER_D_FIELD_G );
		rtp_midi_field_tree = proto_item_add_subtree( ti, ett_rtp_midi_sj_chapter_d_field_g );

		proto_tree_add_item ( rtp_midi_field_tree, hf_rtp_midi_sj_chapter_d_tune_sflag, tvb, offset, 1, FALSE );
		proto_tree_add_item ( rtp_midi_field_tree, hf_rtp_midi_sj_chapter_d_tune_count, tvb, offset, 1, FALSE );

		offset++;
		consumed++;
	}

	/* do we have Song select field? */
	if ( header & RTP_MIDI_SJ_CHAPTER_D_FLAG_H ) {
		/* Can we get the data for the Reset chapter? */
		if ( ( offset + 1 ) > tvb->length ) {
			return -1;
		}

		ti = proto_tree_add_text( rtp_midi_sj_chapter_tree, tvb, offset, 1, RTP_MIDI_TREE_NAME_SJ_CHAPTER_D_FIELD_H );
		rtp_midi_field_tree = proto_item_add_subtree( ti, ett_rtp_midi_sj_chapter_d_field_h );

		proto_tree_add_item ( rtp_midi_field_tree, hf_rtp_midi_sj_chapter_d_song_sel_sflag, tvb, offset, 1, FALSE );
		proto_tree_add_item ( rtp_midi_field_tree, hf_rtp_midi_sj_chapter_d_song_sel_value, tvb, offset, 1, FALSE );

		offset++;
		consumed++;
	}

	/* do we have 0xF4 field? */
	if ( header & RTP_MIDI_SJ_CHAPTER_D_FLAG_J ) {
		ext_consumed = decode_sj_chapter_d_f4( tvb, pinfo, rtp_midi_sj_chapter_tree, offset );
		if ( ext_consumed < 0 ) {
			return ext_consumed;
		}
		consumed += ext_consumed;
		offset += ext_consumed;
	}

	/* do we have 0xF5 field? */
	if ( header & RTP_MIDI_SJ_CHAPTER_D_FLAG_K ) {
		ext_consumed = decode_sj_chapter_d_f5( tvb, pinfo, rtp_midi_sj_chapter_tree, offset );
		if ( ext_consumed < 0 ) {
			return ext_consumed;
		}
		consumed += ext_consumed;
		offset += ext_consumed;
	}

	/* do we have 0xF9 field? */
	if ( header & RTP_MIDI_SJ_CHAPTER_D_FLAG_Y ) {
		ext_consumed = decode_sj_chapter_d_f9( tvb, pinfo, rtp_midi_sj_chapter_tree, offset );
		if ( ext_consumed < 0 ) {
			return ext_consumed;
		}
		consumed += ext_consumed;
		offset += ext_consumed;
	}

	/* do we have 0xFD field? */
	if ( header & RTP_MIDI_SJ_CHAPTER_D_FLAG_Z ) {
		ext_consumed = decode_sj_chapter_d_fd( tvb, pinfo, rtp_midi_sj_chapter_tree, offset );
		if ( ext_consumed < 0 ) {
			return ext_consumed;
		}
		consumed += ext_consumed;
		offset += ext_consumed;
	}


	/* now we know the complete length and set it. */
	proto_item_set_len( tix, consumed );
	return consumed;
}

/*
 * Here the chapter Q of the channel-journal is decoded.
 */
static int
decode_sj_chapter_q( tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned int offset ) {
	return -1;
}

/*
 * Here the chapter F of the channel-journal is decoded.
 */
static int
decode_sj_chapter_f( tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned int offset ) {
	return -1;
}

/*
 * Here the chapter X of the channel-journal is decoded.
 */
static int
decode_sj_chapter_x( tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned int offset ) {
	return -1;
}



/*
 * Here the system-journal is decoded.
 */
static int
decode_system_journal( tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned int offset ) {
	proto_item			*ti				= NULL;
	proto_tree			*rtp_midi_sysjournal_tree	= NULL;
	proto_tree			*rtp_midi_sj_chapters_tree	= NULL;
	proto_tree			*rtp_midi_sj_chapter_tree	= NULL;
	int				consumed = 0;
	int				ext_consumed = 0;
	guint16				sysjourlen;
	guint16				systemflags;



	/* Can we read this channel-journals header? */
	if ( ( offset + 2 ) > tvb->length ) {
		return -1;
	}

	systemflags = tvb_get_ntohs( tvb, offset );
	sysjourlen = systemflags & RTP_MIDI_SJ_MASK_LENGTH;

	ti = proto_tree_add_text( tree, tvb, offset, sysjourlen, RTP_MIDI_TREE_NAME_SYSTEM_JOURNAL );
	rtp_midi_sysjournal_tree = proto_item_add_subtree( ti, ett_rtp_midi_systemjournal );

	proto_tree_add_item (rtp_midi_sysjournal_tree, hf_rtp_midi_sysjour_toc_s, tvb, offset, 2, FALSE );
	proto_tree_add_item (rtp_midi_sysjournal_tree, hf_rtp_midi_sysjour_toc_d, tvb, offset, 2, FALSE );
	proto_tree_add_item (rtp_midi_sysjournal_tree, hf_rtp_midi_sysjour_toc_v, tvb, offset, 2, FALSE );
	proto_tree_add_item (rtp_midi_sysjournal_tree, hf_rtp_midi_sysjour_toc_q, tvb, offset, 2, FALSE );
	proto_tree_add_item (rtp_midi_sysjournal_tree, hf_rtp_midi_sysjour_toc_f, tvb, offset, 2, FALSE );
	proto_tree_add_item (rtp_midi_sysjournal_tree, hf_rtp_midi_sysjour_toc_x, tvb, offset, 2, FALSE );
	proto_tree_add_item (rtp_midi_sysjournal_tree, hf_rtp_midi_sysjour_len, tvb, offset, 2, FALSE );

	offset += 2;
	consumed += 2;

	ti = proto_tree_add_text( rtp_midi_sysjournal_tree, tvb, offset, sysjourlen - 2, RTP_MIDI_TREE_NAME_SYSTEM_CHAPTERS );
	rtp_midi_sj_chapters_tree = proto_item_add_subtree( ti, ett_rtp_midi_systemchapters );

	/* Do we have a simple system commands chapter? */
	if ( systemflags & RTP_MIDI_SJ_FLAG_D ) {
		ext_consumed = decode_sj_chapter_d( tvb, pinfo, rtp_midi_sj_chapters_tree, offset );
		if ( ext_consumed < 0 ) {
			return ext_consumed;
		}
		consumed += ext_consumed;
		offset += ext_consumed;
	}

	/* Do we have a active sensing chapter? */
	if ( systemflags & RTP_MIDI_SJ_FLAG_V ) {
		/* Can we get the data for the Active Sense chapter? */
		if ( ( offset + 1 ) > tvb->length ) {
			return -1;
		}

		ti = proto_tree_add_text( rtp_midi_sj_chapters_tree, tvb, offset, 1, RTP_MIDI_TREE_NAME_SJ_CHAPTER_V );
		rtp_midi_sj_chapter_tree = proto_item_add_subtree( ti, ett_rtp_midi_sj_chapter_v );

		proto_tree_add_item ( rtp_midi_sj_chapter_tree, hf_rtp_midi_sj_chapter_v_sflag, tvb, offset, 1, FALSE );
		proto_tree_add_item ( rtp_midi_sj_chapter_tree, hf_rtp_midi_sj_chapter_v_count, tvb, offset, 1, FALSE );

		offset++;
		consumed++;
	}

	/* Do we have a sequencer state commands chapter? */
	if ( systemflags & RTP_MIDI_SJ_FLAG_Q ) {
		ext_consumed = decode_sj_chapter_q( tvb, pinfo, rtp_midi_sj_chapters_tree, offset );
		if ( ext_consumed < 0 ) {
			return ext_consumed;
		}
		consumed += ext_consumed;
		offset += ext_consumed;
	}

	/* Do we have a MTC chapter? */
	if ( systemflags & RTP_MIDI_SJ_FLAG_F ) {
		ext_consumed = decode_sj_chapter_f( tvb, pinfo, rtp_midi_sj_chapters_tree, offset );
		if ( ext_consumed < 0 ) {
			return ext_consumed;
		}
		consumed += ext_consumed;
		offset += ext_consumed;
	}

	/* Do we have a Sysex chapter? */
	if ( systemflags & RTP_MIDI_SJ_FLAG_X ) {
		ext_consumed = decode_sj_chapter_x( tvb, pinfo, rtp_midi_sj_chapters_tree, offset );
		if ( ext_consumed < 0 ) {
			return ext_consumed;
		}
		consumed += ext_consumed;
		offset += ext_consumed;
	}


	/* Make sanity check for consumed data vs. stated length of system journal */
	if ( consumed != sysjourlen ) {
		return -1;
	}
	return consumed;
}






/*
 * This is the main dissector function for the RTP-MIDI protocol
 * It decodes the wrapper of the command-section and punts
 * decoding of delta-time and the actual MIDI-command itself to
 * external decoders.  Afterwards the journal-section is decoded.
 */

static void
dissect_rtp_midi( tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree )
{
	proto_item			*ti				= NULL;
	proto_item			*command_item			= NULL;
	proto_item			*journal_item			= NULL;
	proto_tree			*rtp_midi_tree			= NULL;
	proto_tree			*rtp_midi_commands_tree		= NULL;
	proto_tree			*rtp_midi_journal_tree		= NULL;
	proto_tree			*rtp_midi_chanjournals_tree	= NULL;
	unsigned int			 offset				= 0;
	struct _rtp_conversation_info	*p_conv_data			= NULL;

	guint8				flags;				/* used for command-section and journal-section*/
	guint8				octet;
	unsigned int			cmd_len;
	unsigned int			cmd_count;
	guint8				runningstatus;

	unsigned int			consumed;
	unsigned int			rsoffset;

	guint8				totchan;

	unsigned int			i;

	col_set_str( pinfo->cinfo, COL_PROTOCOL, RTP_MIDI_DISSECTOR_SHORTNAME );
	col_clear( pinfo->cinfo, COL_INFO );

	/*
	 * MIDI command section
	 */

	/* RTP-MIDI starts with 4 bits of flags... */
	flags = tvb_get_guint8( tvb, offset );

	ti = proto_tree_add_item( tree, proto_rtp_midi, tvb, 0, -1, FALSE );
	rtp_midi_tree = proto_item_add_subtree( ti, ett_rtp_midi);

	/* Decode those four flags */
	proto_tree_add_item ( rtp_midi_tree, hf_rtp_midi_bflag, tvb, offset, 1, FALSE );
	proto_tree_add_item ( rtp_midi_tree, hf_rtp_midi_jflag, tvb, offset, 1, FALSE );
	proto_tree_add_item ( rtp_midi_tree, hf_rtp_midi_zflag, tvb, offset, 1, FALSE );
	proto_tree_add_item ( rtp_midi_tree, hf_rtp_midi_pflag, tvb, offset, 1, FALSE );

	/* ...followed by a length-field of at least 4 bits */
	cmd_len = flags & RTP_MIDI_CS_MASK_SHORTLEN;

	/* see if we have small or large len-field */
	if (flags & RTP_MIDI_CS_FLAG_B)  {
		octet = tvb_get_guint8( tvb, offset+1 );
		cmd_len = ( cmd_len << 8 ) | octet;
		proto_tree_add_item ( rtp_midi_tree, hf_rtp_midi_longlen, tvb, offset, 2, FALSE );
		offset += 2;
	} else {
		proto_tree_add_item ( rtp_midi_tree, hf_rtp_midi_shortlen, tvb, offset, 1, FALSE );
			offset++;
	}

	/* if we have a command-section -> dissect it */
	if ( cmd_len ) {
		command_item = proto_tree_add_text( rtp_midi_tree, tvb, offset, cmd_len, RTP_MIDI_TREE_NAME_COMMAND );
		rtp_midi_commands_tree = proto_item_add_subtree( command_item, ett_rtp_midi_commands );

		/* if the reported command-length larger than data found in packet -> error */
		if ( ( offset + cmd_len ) > tvb->length ) {
			THROW( ReportedBoundsError );
			return;
		}

		/* No commands decoded yet */
		cmd_count=0;

		/* RTP-MIDI-pdus always start with no running status */
		runningstatus=0;

		/* Multiple MIDI-commands might follow - the exact number can only be discovered by really decoding the commands! */
		while ( cmd_len) {

			/* for the first command we only have a delta-time if Z-Flag is set */
			if ( ( cmd_count ) || ( flags & RTP_MIDI_CS_FLAG_Z ) ) {

				/* Decode a delta-time - if 0 is returned something went wrong */
				consumed=decodetime( tvb, pinfo, rtp_midi_commands_tree, offset, cmd_len );
				if ( -1 == consumed ) {
					THROW( ReportedBoundsError );
					return;
				}

				/* seek to next command and set remaining length */
				offset += consumed;
				cmd_len -= consumed;
			}

			/* Only decode MIDI-command if there is any data left - it is valid to only have delta-time! */
			if ( cmd_len ) {

				/* Decode a MIDI-command - if 0 is returned something went wrong */
				consumed = decodemidi( tvb, pinfo, rtp_midi_commands_tree, cmd_count, offset, cmd_len, &runningstatus, &rsoffset );
				if ( -1 == consumed ) {
					THROW( ReportedBoundsError );
					return;
				}

				/* seek to next delta-time and set remaining length */
				offset += consumed;
				cmd_len -= consumed;

				/* as we have successfully decoded another command, increment count */
				cmd_count++;
			}
		}
	}
	/*
	 * Journal section
	 */

	/* if we have a journal-section -> dissect it */
	if ( flags & RTP_MIDI_CS_FLAG_J ) {

		journal_item = proto_tree_add_text( rtp_midi_tree, tvb, offset, -1, RTP_MIDI_TREE_NAME_JOURNAL );
		rtp_midi_journal_tree = proto_item_add_subtree( journal_item, ett_rtp_midi_journal );

		/* lets get the main flags from the recovery journal header */
		flags = tvb_get_guint8( tvb, offset );

		proto_tree_add_item (rtp_midi_journal_tree, hf_rtp_midi_sflag, tvb, offset, 1, FALSE );
		proto_tree_add_item (rtp_midi_journal_tree, hf_rtp_midi_yflag, tvb, offset, 1, FALSE );
		proto_tree_add_item (rtp_midi_journal_tree, hf_rtp_midi_aflag, tvb, offset, 1, FALSE );
		proto_tree_add_item (rtp_midi_journal_tree, hf_rtp_midi_hflag, tvb, offset, 1, FALSE );

		/* At the same place we find the total channels encoded in the channel journal */
		totchan = flags & RTP_MIDI_JS_MASK_TOTALCHANNELS;
		proto_tree_add_item ( rtp_midi_journal_tree, hf_rtp_midi_totchan, tvb, offset, 1, 0);
		offset++;

		/* the checkpoint-sequence-number can be used to see if the recovery journal covers all lost events */
		proto_tree_add_item ( rtp_midi_journal_tree, hf_rtp_midi_check_seq_num, tvb, offset, 2, FALSE );
		offset += 2;


		/* do we have system journal? */
		if ( flags & RTP_MIDI_JS_FLAG_Y ) {
			/* first we need to get the flags & length from the system-journal */
			consumed = decode_system_journal( tvb, pinfo, rtp_midi_journal_tree, offset );

			if ( -1 == consumed ) {
				THROW( ReportedBoundsError );
				return;
			}

			/* seek to optional channel-journals-section */
			offset += consumed;

		}

		/* do we have channel journal(s)? */
		if ( flags & RTP_MIDI_JS_FLAG_A ) {
			ti = proto_tree_add_text( rtp_midi_journal_tree, tvb, offset, -1, RTP_MIDI_TREE_NAME_CHANNEL_JOURNAL );
			rtp_midi_chanjournals_tree = proto_item_add_subtree( ti, ett_rtp_midi_channeljournals );

			/* iterate through all the channels specified in header */
			for ( i = 0; i <= totchan; i++ ) {

				consumed = decode_channel_journal( tvb, pinfo, rtp_midi_chanjournals_tree, offset );

				if ( -1 == consumed ) {
					THROW( ReportedBoundsError );
					return;
				}

				/* seek to next channel-journal */
				offset += consumed;

			}
		}
	}
}




void
proto_register_rtp_midi( void )
{
	module_t *rtp_midi_module;

	proto_rtp_midi = proto_register_protocol( RTP_MIDI_DISSECTOR_NAME, RTP_MIDI_DISSECTOR_SHORTNAME, RTP_MIDI_DISSECTOR_ABBREVIATION );
	proto_register_field_array( proto_rtp_midi, hf, array_length( hf ) );
	proto_register_subtree_array( ett, array_length( ett ) );

	/* Register preferences - see reference to AppleMIDI at top of this file for explanation */
	rtp_midi_module = prefs_register_protocol ( proto_rtp_midi, proto_reg_handoff_rtp_midi );
	prefs_register_uint_preference ( rtp_midi_module, "midi_payload_type_value", "Payload Type for RFC 4695 RTP-MIDI", "This is the value of the Payload Type field that specifies RTP-MIDI", 10, &rtp_midi_payload_type_value );
	register_dissector( "rtpmidi", dissect_rtp_midi, proto_rtp_midi );
}



void
proto_reg_handoff_rtp_midi( void )
{
	static dissector_handle_t	rtp_midi_handle;
	static int			rtp_midi_prefs_initialized = FALSE;


	if ( !rtp_midi_prefs_initialized ) {
		rtp_midi_handle = find_dissector( RTP_MIDI_DISSECTOR_ABBREVIATION );
		dissector_add_string("rtp_dyn_payload_type", "rtp-midi", rtp_midi_handle);
		rtp_midi_prefs_initialized = TRUE;
	}
	else {
		dissector_delete( "rtp.pt", saved_payload_type_value, rtp_midi_handle );
	}

	saved_payload_type_value = rtp_midi_payload_type_value;
	/* rtp_event_payload_type_value is set from preferences */

	dissector_add( "rtp.pt", saved_payload_type_value, rtp_midi_handle );
	dissector_add_string( "rtp_dyn_payload_type", "midi", rtp_midi_handle );
}
